import uuid
from dataclasses import dataclass
from typing import Any, Optional, Union


@dataclass
class FileAttachment:
    path: str
    id: Optional[str] = None
    name: Optional[str] = None
    size: Optional[int] = None
    mime_type: Optional[str] = None
    tags: Optional[dict[str, str]] = None
    metadata: Optional[dict[str, Any]] = None
    timestamp: Optional[int] = None

    def __post_init__(self) -> None:
        if self.id is None:
            self.id = str(uuid.uuid4())

    def to_dict(self) -> dict[str, Any]:
        """
        Convert the file attachment to a dictionary.

        Returns:
            A dictionary containing the file attachment.
        """
        return {
            "id": self.id,
            "type": "file",
            "name": self.name,
            "path": self.path,
            "mime_type": self.mime_type,
            "tags": self.tags,
            "metadata": self.metadata,
            "timestamp": self.timestamp,
        }


@dataclass
class FileDataAttachment:
    data: bytes
    id: Optional[str] = None
    name: Optional[str] = None
    size: Optional[int] = None
    mime_type: Optional[str] = None
    tags: Optional[dict[str, str]] = None
    metadata: Optional[dict[str, Any]] = None
    timestamp: Optional[int] = None

    def __post_init__(self) -> None:
        """
        Initialize the file data attachment.
        """
        if self.id is None:
            self.id = str(uuid.uuid4())

    def to_dict(self) -> dict[str, Any]:
        """
        Convert the file data attachment to a dictionary.

        Returns:
            A dictionary containing the file data attachment.
        """
        return {
            "id": self.id,
            "type": "file_data",
            "name": self.name,
            "data": self.data,
            "mime_type": self.mime_type,
            "tags": self.tags,
            "metadata": self.metadata,
            "timestamp": self.timestamp,
        }


@dataclass
class UrlAttachment:
    url: str
    id: Optional[str] = None
    name: Optional[str] = None
    size: Optional[int] = None
    mime_type: Optional[str] = None
    tags: Optional[dict[str, str]] = None
    metadata: Optional[dict[str, Any]] = None
    timestamp: Optional[int] = None

    def __post_init__(self) -> None:
        """
        Initialize the url attachment.
        """
        if self.id is None:
            self.id = str(uuid.uuid4())

    def to_dict(self) -> dict[str, Any]:
        """
        Convert the url attachment to a dictionary.

        Returns:
            A dictionary containing the url attachment.
        """
        return {
            "id": self.id,
            "type": "url",
            "name": self.name,
            "url": self.url,
            "mime_type": self.mime_type,
            "tags": self.tags,
            "metadata": self.metadata,
            "timestamp": self.timestamp,
        }


Attachment = Union[FileAttachment, FileDataAttachment, UrlAttachment]



"""
Configuration class for Maxim Logger.

This class holds the configuration settings for the Logger, including
the logger ID, auto-flush setting, and flush interval.

Attributes:
    id (str): The unique identifier for the logger.
    auto_flush (bool): Whether to automatically flush logs. Defaults to True.
    flush_interval (int): The interval (in seconds) at which to flush logs when auto_flush is True. Defaults to 10 seconds.
"""

import threading
from typing import Any, Dict, Optional, TypedDict, Union

from typing_extensions import deprecated

from ..logger.components.types import Entity
from ..scribe import scribe
from .components import (
    Error,
    ErrorConfig,
    Feedback,
    FeedbackDict,
    FileAttachment,
    FileDataAttachment,
    Generation,
    GenerationConfig,
    GenerationConfigDict,
    GenerationError,
    GenerationRequestMessage,
    Retrieval,
    RetrievalConfig,
    RetrievalConfigDict,
    Session,
    SessionConfig,
    SessionConfigDict,
    Span,
    SpanConfig,
    SpanConfigDict,
    ToolCall,
    ToolCallConfig,
    ToolCallConfigDict,
    ToolCallError,
    ToolCallErrorDict,
    Trace,
    TraceConfig,
    TraceConfigDict,
    UrlAttachment,
)
from .components.base import EvaluateContainer
from .writer import LogWriter, LogWriterConfig


@deprecated(
    "This class will be removed in a future version. Use LoggerConfigDict instead."
)
class LoggerConfig:
    """
    Configuration class for Maxim Logger.

    This class holds the configuration settings for the Logger, including
    the logger ID, auto-flush setting, and flush interval.

    Attributes:
        id (str): The unique identifier for the logger.
        auto_flush (bool): Whether to automatically flush logs.
        flush_interval (int): The interval (in seconds) at which to flush logs when auto_flush is True.
    """

    def __init__(self, id: str, auto_flush=True, flush_interval=10) -> None:
        self.id = id
        self.auto_flush = auto_flush
        self.flush_interval = flush_interval


class LoggerConfigDict(TypedDict, total=False):
    id: str
    auto_flush: bool
    flush_interval: int


def get_logger_config_dict(
    config: Union[LoggerConfig, LoggerConfigDict],
) -> LoggerConfigDict:
    if isinstance(config, LoggerConfig):
        return LoggerConfigDict(
            id=config.id,
            auto_flush=config.auto_flush,
            flush_interval=config.flush_interval,
        )
    else:
        return config


class Logger:
    """
    A class representing a logger for the Maxim SDK.

    This logger provides methods for creating sessions, traces, and various logging components
    such as spans, generations, retrievals, and tool calls. It uses a LogWriter to handle the
    actual logging operations.

    Attributes:
        _id (str): The unique identifier for this logger instance.
        raise_exceptions (bool): Whether to raise exceptions during logging operations.
        is_debug (bool): Whether debug logging is enabled.
        writer (LogWriter): The LogWriter instance used for actual logging operations.
    """

    def __init__(
        self,
        config: LoggerConfigDict,
        api_key: str,
        base_url: str,
        is_debug=False,
        raise_exceptions=False,
    ) -> None:
        """
        Initializes the logger with the given configuration.

        Args:
            config (LoggerConfig): The configuration for the logger.
            api_key (str): The API key for the logger.
            base_url (str): The base URL for the logger.
            is_debug (bool, optional): Whether to enable debug logging. Defaults to False.
            raise_exceptions (bool, optional): Whether to raise exceptions. Defaults to False.
        """
        repo_id = config.get("id", None)
        if repo_id is None:
            raise ValueError("Logger must be initialized with id of the logger")
        self._id = repo_id
        self.raise_exceptions = raise_exceptions
        self.is_debug = is_debug
        writer_config = LogWriterConfig(
            auto_flush="auto_flush" in config and config["auto_flush"] or True,
            flush_interval="flush_interval" in config
            and config["flush_interval"]
            or 10,
            base_url=base_url,
            api_key=api_key,
            is_debug=is_debug,
            repository_id=repo_id,
            raise_exceptions=raise_exceptions,
        )
        self.writer = LogWriter(writer_config)
        scribe().debug("[MaximSDK] Logger initialized")

    def session(self, config: Union[SessionConfig, SessionConfigDict]) -> Session:
        """
        Creates a new session with the given configuration.

        Args:
            config (SessionConfig): The configuration for the new session.

        Returns:
            Session: The newly created session.
        """
        return Session(config, self.writer)

    def trace(self, config: Union[TraceConfig, TraceConfigDict]) -> Trace:
        """
        Creates a new trace with the given configuration.

        Args:
            config (TraceConfig): The configuration for the new trace.

        Returns:
            Trace: The newly created trace.
        """
        return Trace(config, self.writer)

    # Session methods
    def session_add_tag(self, session_id: str, key: str, value: str):
        """
        Adds a tag to the session.

        Args:
            session_id (str): The ID of the session.
            key (str): The key of the tag.
            value (str): The value of the tag.
        """
        Session.add_tag_(self.writer, session_id, key, value)

    def session_end(self, session_id: str):
        """
        Ends the session.

        Args:
            session_id (str): The ID of the session.
        """
        Session.end_(self.writer, session_id)

    def session_event(self, session_id: str, event_id: str, event: str, data: Any):
        """
        Adds an event to the session.

        Args:
            session_id (str): The ID of the session.
            event_id (str): The ID of the event.
            event (str): The name of the event.
            data (Any): The data associated with the event.
        """
        Session.event_(self.writer, session_id, event_id, event, data)

    @deprecated(
        "This method will be removed in a future version. Use session_add_feedback instead."
    )
    def session_feedback(
        self, session_id: str, feedback: Union[Feedback, FeedbackDict]
    ):
        """
        Adds a feedback to the session.

        Args:
            session_id (str): The ID of the session.
            feedback (Feedback): The feedback to add.
        """
        Session.feedback_(self.writer, session_id, feedback)

    def session_add_attachment(self, session_id: str, attachment: Union[FileAttachment, FileDataAttachment, UrlAttachment]):
        """
        Adds an attachment to the session.
        """
        Session.add_attachment_(self.writer, session_id, attachment)

    def session_add_feedback(self, session_id: str, feedback: FeedbackDict):
        """
        Adds a feedback to the session.
        """
        Session.feedback_(self.writer, session_id, feedback)

    @deprecated(
        "This method will be removed in a future version. Use session_add_trace instead."
    )
    def session_trace(self, session_id: str, config: TraceConfig) -> Trace:
        """
        Adds a trace to the session.

        Args:
            session_id (str): The ID of the session.
            config (TraceConfig): The configuration for the trace.

        Returns:
            Trace: The newly created trace.
        """
        return Session.trace_(self.writer, session_id, config)

    def session_add_trace(
        self, session_id: str, config: Union[TraceConfig, TraceConfigDict]
    ) -> Trace:
        """
        Adds a trace to the session.

        Args:
            session_id (str): The ID of the session.
            config (TraceConfig): The configuration for the trace.

        Returns:
            Trace: The newly created trace.
        """
        return Session.trace_(self.writer, session_id, config)

    # Trace methods
    @deprecated(
        "This method will be removed in a future version. Use trace_add_generation instead."
    )
    def trace_generation(self, trace_id: str, config: GenerationConfig) -> Generation:
        """
        Adds a generation to the trace.

        Args:
            trace_id (str): The ID of the trace.
            config (GenerationConfig): The configuration for the generation.

        Returns:
            Generation: The newly created generation.
        """
        return Trace.generation_(self.writer, trace_id, config)

    def trace_add_generation(
        self, trace_id: str, config: Union[GenerationConfig, GenerationConfigDict]
    ) -> Generation:
        """
        Adds a generation to the trace.

        Args:
            trace_id (str): The ID of the trace.
            config (GenerationConfig): The configuration for the generation.

        Returns:
            Generation: The newly created generation.
        """
        return Trace.generation_(self.writer, trace_id, config)

    @deprecated(
        "This method will be removed in a future version. Use trace_add_retrieval instead."
    )
    def trace_retrieval(
        self, trace_id: str, config: Union[RetrievalConfig, RetrievalConfigDict]
    ) -> Retrieval:
        """
        Adds a retrieval to the trace.

        Args:
            trace_id (str): The ID of the trace.
            config (RetrievalConfig): The configuration for the retrieval.

        Returns:
            Retrieval: The newly created retrieval.
        """
        return Trace.retrieval_(self.writer, trace_id, config)

    def trace_add_retrieval(
        self, trace_id: str, config: Union[RetrievalConfig, RetrievalConfigDict]
    ) -> Retrieval:
        """
        Adds a retrieval to the trace.

        Args:
            trace_id (str): The ID of the trace.
            config (RetrievalConfig): The configuration for the retrieval.

        Returns:
            Retrieval: The newly created retrieval.
        """
        return Trace.retrieval_(self.writer, trace_id, config)

    @deprecated(
        "This method will be removed in a future version. Use trace_add_span instead."
    )
    def trace_span(
        self, trace_id: str, config: Union[SpanConfig, SpanConfigDict]
    ) -> Span:
        """
        Adds a span to the trace.

        Args:
            trace_id (str): The ID of the trace.
            config (SpanConfig): The configuration for the span.

        Returns:
            Span: The newly created span.
        """
        return Trace.span_(self.writer, trace_id, config)

    def trace_add_span(
        self, trace_id: str, config: Union[SpanConfig, SpanConfigDict]
    ) -> Span:
        """
        Adds a span to the trace.

        Args:
            trace_id (str): The ID of the trace.
            config (SpanConfig): The configuration for the span.

        Returns:
            Span: The newly created span.
        """
        return Trace.span_(self.writer, trace_id, config)

    def trace_add_error(self, trace_id: str, config: ErrorConfig) -> Error:
        """
        Adds an error to the trace.
        """
        return Trace.error_(self.writer, trace_id, config)

    def trace_add_tag(self, trace_id: str, key: str, value: str):
        """
        Adds a tag to the trace.

        Args:
            trace_id (str): The ID of the trace.
            key (str): The key of the tag.
            value (str): The value of the tag.
        """
        Trace.add_tag_(self.writer, trace_id, key, value)

    def trace_add_tool_call(
        self, trace_id: str, config: Union[ToolCallConfig, ToolCallConfigDict]
    ) -> ToolCall:
        """
        Adds a tool call to the trace.

        Args:
            trace_id (str): The ID of the trace.
            config (ToolCallConfig): The configuration for the tool call.

        Returns:
            ToolCall: The newly created tool call.
        """
        return Trace.tool_call_(self.writer, trace_id, config)

    def trace_evaluate(self, trace_id: str) -> EvaluateContainer:
        return Trace._evaluate_(self.writer, Entity.TRACE, trace_id)

    @deprecated(
        "This method will be removed in a future version. Use trace_add_event instead."
    )
    def trace_event(
        self,
        trace_id: str,
        event_id: str,
        event: str,
        tags: Optional[Dict[str, str]] = None,
        metadata: Optional[Dict[str, Any]] = None,
    ):
        """
        Adds an event to the trace.

        Args:
            trace_id (str): The ID of the trace.
            event_id (str): The ID of the event.
            event (str): The name of the event.
            tags (Optional[Dict[str, str]]): The tags associated with the event.
        """
        Trace.event_(self.writer, trace_id, event_id, event, tags, metadata)

    def trace_add_event(
        self,
        trace_id: str,
        event_id: str,
        event: str,
        tags: Optional[Dict[str, str]] = None,
        metadata: Optional[Dict[str, Any]] = None,
    ):
        """
        Adds an event to the trace.
        """
        Trace.event_(self.writer, trace_id, event_id, event, tags, metadata)

    def trace_set_input(self, trace_id: str, input: str):
        """
        Sets the input for the trace.

        Args:
            trace_id (str): The ID of the trace.
            input (str): The input for the trace.
        """
        Trace.set_input_(self.writer, trace_id, input)

    def trace_set_output(self, trace_id: str, output: str):
        """
        Sets the output for the trace.

        Args:
            trace_id (str): The ID of the trace.
            output (str): The output for the trace.
        """
        Trace.set_output_(self.writer, trace_id, output)

    @deprecated(
        "This method will be removed in a future version. Use trace_add_feedback instead."
    )
    def trace_feedback(self, trace_id: str, feedback: Feedback):
        """
        Adds a feedback to the trace.

        Args:
            trace_id (str): The ID of the trace.
            feedback (Feedback): The feedback to add.
        """
        Trace.feedback_(self.writer, trace_id, feedback)

    def trace_add_feedback(self, trace_id: str, feedback: FeedbackDict):
        """
        Adds a feedback to the trace.
        """
        Trace.feedback_(self.writer, trace_id, feedback)

    def trace_add_metadata(self, trace_id: str, metadata: Dict[str, Any]):
        """
        Adds metadata to the trace.

        Args:
            trace_id (str): The ID of the trace.
            metadata (Dict[str, Any]): The metadata to add.
        """
        Trace.add_metadata_(self.writer, Entity.TRACE, trace_id, metadata)

    def trace_add_attachment(
        self,
        trace_id: str,
        attachment: Union[FileAttachment, FileDataAttachment, UrlAttachment],
    ):
        """
        Adds an attachment to the trace.
        """
        Trace.add_attachment_(self.writer, trace_id, attachment)

    def trace_end(self, trace_id: str):
        """
        Ends the trace.

        Args:
            trace_id (str): The ID of the trace.
        """
        Trace.end_(self.writer, trace_id)

    # Generation methods
    def generation_set_model(self, generation_id: str, model: str):
        """
        Sets the model for the generation.

        Args:
            generation_id (str): The ID of the generation.
            model (str): The model for the generation.
        """
        Generation.set_model_(self.writer, generation_id, model)

    def generation_set_provider(self, generation_id: str, provider: str):
        """
        Sets the provider for the generation.
        """
        Generation.set_provider_(self.writer, generation_id, provider)

    def generation_add_message(
        self, generation_id: str, message: GenerationRequestMessage
    ):
        """
        Adds a message to the generation.

        Args:
            generation_id (str): The ID of the generation.
            message (Any): The OpenAI chat message to add.
        """
        Generation.add_message_(self.writer, generation_id, message)

    def generation_set_model_parameters(
        self, generation_id: str, model_parameters: Dict[str, Any]
    ):
        """
        Sets the model parameters for the generation.

        Args:
            generation_id (str): The ID of the generation.
            model_parameters (dict): The model parameters for the generation.
        """
        Generation.set_model_parameters_(self.writer, generation_id, model_parameters)

    def generation_result(self, generation_id: str, result: Any):
        """
        Sets the result for the generation.

        Args:
            generation_id (str): The ID of the generation.
            result (Any): The result for the generation.
        """
        Generation.result_(self.writer, generation_id, result)

    def generation_add_attachment(
        self,
        generation_id: str,
        attachment: Union[FileAttachment, FileDataAttachment, UrlAttachment],
    ):
        """
        Adds an attachment to the generation.
        """
        Generation.add_attachment_(self.writer, generation_id, attachment)

    def generation_end(self, generation_id: str):
        """
        Ends the generation.

        Args:
            generation_id (str): The ID of the generation.
        """
        Generation.end_(self.writer, generation_id)

    def generation_error(self, generation_id: str, error: GenerationError):
        """
        Sets the error for the generation.

        Args:
            generation_id (str): The ID of the generation.
            error (GenerationError): The error for the generation.
        """
        Generation.error_(self.writer, generation_id, error)

    def generation_add_metadata(self, generation_id: str, metadata: Dict[str, Any]):
        Generation.add_metadata_(
            self.writer, Entity.GENERATION, generation_id, metadata
        )

    def generation_evaluate(self, generation_id: str) -> EvaluateContainer:
        return Generation._evaluate_(self.writer, Entity.GENERATION, generation_id)

    # Span methods
    @deprecated(
        "This method will be removed in a future version. Use span_add_generation instead."
    )
    def span_generation(
        self, span_id: str, config: Union[GenerationConfig, GenerationConfigDict]
    ) -> Generation:
        """
        Adds a generation to the span.

        Args:
            span_id (str): The ID of the span.
            config (GenerationConfig): The configuration for the generation.

        Returns:
            Generation: The newly created generation.
        """
        return Span.generation_(self.writer, span_id, config)

    def span_add_generation(
        self, span_id: str, config: Union[GenerationConfig, GenerationConfigDict]
    ) -> Generation:
        """
        Adds a generation to the span.

        Args:
            span_id (str): The ID of the span.
            config (GenerationConfig): The configuration for the generation.

        Returns:
            Generation: The newly created generation.
        """
        return Span.generation_(self.writer, span_id, config)

    def span_add_error(self, span_id: str, config: ErrorConfig) -> Error:
        """
        Adds an error to the span.
        """
        return Span.error_(self.writer, span_id, config)

    @deprecated(
        "This method will be removed in a future version. Use span_add_retrieval instead."
    )
    def span_retrieval(
        self, span_id: str, config: Union[RetrievalConfig, RetrievalConfigDict]
    ) -> Retrieval:
        """
        Adds a retrieval to the span.

        Args:
            span_id (str): The ID of the span.
            config (RetrievalConfig): The configuration for the retrieval.

        Returns:
            Retrieval: The newly created retrieval.
        """
        return Span.retrieval_(self.writer, span_id, config)

    def span_add_retrieval(
        self, span_id: str, config: Union[RetrievalConfig, RetrievalConfigDict]
    ) -> Retrieval:
        """
        Adds a retrieval to the span.

        Args:
            span_id (str): The ID of the span.
            config (RetrievalConfig): The configuration for the retrieval.

        Returns:
            Retrieval: The newly created retrieval.
        """
        return Span.retrieval_(self.writer, span_id, config)

    def span_add_tool_call(
        self, span_id: str, config: Union[ToolCallConfig, ToolCallConfigDict]
    ) -> ToolCall:
        """
        Adds a tool call to the span.

        Args:
            span_id (str): The ID of the span.
            config (ToolCallConfig): The configuration for the tool call.

        Returns:
            ToolCall: The newly created tool call.
        """
        return Span.tool_call_(self.writer, span_id, config)

    def span_end(self, span_id: str):
        """
        Ends the span.

        Args:
            span_id (str): The ID of the span.
        """
        Span.end_(self.writer, span_id)

    def span_add_tag(self, span_id: str, key: str, value: str):
        """
        Adds a tag to the span.

        Args:
            span_id (str): The ID of the span.
            key (str): The key of the tag.
            value (str): The value of the tag.
        """
        Span.add_tag_(self.writer, span_id, key, value)

    def span_event(
        self,
        span_id: str,
        event_id: str,
        name: str,
        tags: Optional[Dict[str, str]] = None,
        metadata: Optional[Dict[str, Any]] = None,
    ):
        """
        Adds an event to the span.

        Args:
            span_id (str): The ID of the span.
            event_id (str): The ID of the event.
            name (str): The name of the event.
            tags (Optional[Dict[str, str]]): The tags associated with the event.
        """
        Span.event_(self.writer, span_id, event_id, name, tags, metadata)

    def span_add_metadata(self, span_id: str, metadata: Dict[str, Any]):
        """
        Adds metadata to the span.

        Args:
            span_id (str): The ID of the span.
            metadata (Dict[str, Any]): The metadata to add.
        """
        Span.add_metadata_(self.writer, Entity.SPAN, span_id, metadata)

    def span_add_attachment(
        self,
        span_id: str,
        attachment: Union[FileAttachment, FileDataAttachment, UrlAttachment],
    ):
        """
        Adds an attachment to the span.
        """
        Span.add_attachment_(self.writer, span_id, attachment)

    @deprecated(
        "This method will be removed in a future version. Use span_add_sub_span instead."
    )
    def span_span(
        self, span_id: str, config: Union[SpanConfig, SpanConfigDict]
    ) -> Span:
        """
        Adds a span to the span.

        Args:
            span_id (str): The ID of the span.
            config (SpanConfig): The configuration for the sub-span.

        Returns:
            Span: The newly created sub-span.
        """
        return Span.span_(self.writer, span_id, config)

    def span_add_sub_span(
        self, span_id: str, config: Union[SpanConfig, SpanConfigDict]
    ) -> Span:
        """
        Adds a sub-span to the span.

        Args:
            span_id (str): The ID of the span.
            config (SpanConfig): The configuration for the sub-span.

        Returns:
            Span: The newly created sub-span.
        """
        return Span.span_(self.writer, span_id, config)

    def span_evaluate(self, span_id: str) -> EvaluateContainer:
        return Span._evaluate_(self.writer, Entity.SPAN, span_id)

    # Retrieval methods
    def retrieval_end(self, retrieval_id: str):
        """
        Ends the retrieval.

        Args:
            retrieval_id (str): The ID of the retrieval.
        """
        Retrieval.end_(self.writer, retrieval_id)

    def retrieval_input(self, retrieval_id: str, query: Any):
        """
        Sets the input for the retrieval.

        Args:
            retrieval_id (str): The ID of the retrieval.
            query (Any): The input for the retrieval.
        """
        Retrieval.input_(self.writer, retrieval_id, query)

    def retrieval_output(self, retrieval_id: str, docs: Any):
        """
        Sets the output for the retrieval.

        Args:
            retrieval_id (str): The ID of the retrieval.
            docs (Any): The output for the retrieval.
        """
        Retrieval.output_(self.writer, retrieval_id, docs)

    def retrieval_add_tag(self, retrieval_id: str, key: str, value: str):
        """
        Adds a tag to the retrieval.

        Args:
            retrieval_id (str): The ID of the retrieval.
            key (str): The key of the tag.
            value (str): The value of the tag.
        """
        Retrieval.add_tag_(self.writer, retrieval_id, key, value)

    def retrieval_add_metadata(self, retrieval_id: str, metadata: Dict[str, Any]):
        Retrieval.add_metadata_(self.writer, Entity.RETRIEVAL, retrieval_id, metadata)

    def retrieval_evaluate(self, retrieval_id: str) -> EvaluateContainer:
        return Retrieval._evaluate_(self.writer, Entity.RETRIEVAL, retrieval_id)

    def retrieval_add_attachment(
        self,
        retrieval_id: str,
        attachment: Union[FileAttachment, FileDataAttachment, UrlAttachment],
    ):
        """
        Adds an attachment to the retrieval.
        """
        Retrieval.add_attachment_(self.writer, retrieval_id, attachment)

    # Tool call methods
    def tool_call_update(self, tool_call_id: str, data: Dict[str, Any]):
        """
        Updates the tool call.

        Args:
            tool_call_id (str): The ID of the tool call.
            data (Dict[str, Any]): The data to update the tool call with.
        """
        ToolCall.update_(self.writer, tool_call_id, data)

    def tool_call_result(self, tool_call_id: str, result: Any):
        """
        Sets the result for the tool call.

        Args:
            tool_call_id (str): The ID of the tool call.
            result (Any): The result for the tool call.
        """
        ToolCall.result_(self.writer, tool_call_id, result)

    def tool_call_error(
        self, tool_call_id: str, error: Union[ToolCallError, ToolCallErrorDict]
    ):
        """
        Sets the error for the tool call.

        Args:
            tool_call_id (str): The ID of the tool call.
            error (ToolCallError): The error for the tool call.
        """
        ToolCall.error_(self.writer, tool_call_id, error)

    def tool_call_add_metadata(self, tool_call_id: str, metadata: Dict[str, Any]):
        """
        Adds metadata to the tool call.

        Args:
            tool_call_id (str): The ID of the tool call.
            metadata (Dict[str, Any]): The metadata to add.
        """
        ToolCall.add_metadata_(self.writer, Entity.TOOL_CALL, tool_call_id, metadata)

    @property
    def id(self):
        """
        Returns the ID of the logger.
        """
        return self._id

    def flush(self):
        """
        Flushes the writer.
        """
        self.writer.flush()

    def cleanup(self, is_sync=False):
        """
        Cleans up the writer.
        """
        self.writer.cleanup(is_sync)



from openai import AsyncOpenAI
from openai.resources.chat import AsyncChat

from ..logger import (
    Logger,
)
from .async_completions import MaximAsyncOpenAIChatCompletions


class MaximAsyncOpenAIChat(AsyncChat):
    def __init__(self, client: AsyncOpenAI, logger: Logger):
        super().__init__(client=client)
        self._logger = logger

    @property
    def completions(self) -> MaximAsyncOpenAIChatCompletions:
        return MaximAsyncOpenAIChatCompletions(self._client, self._logger)



from __future__ import annotations

from typing import Any, Iterable, List, Optional, Union
from uuid import uuid4

import httpx
from anthropic import Anthropic, MessageStreamEvent
from anthropic._types import NOT_GIVEN, Body, Headers, NotGiven, Query
from anthropic.resources.messages import Messages
from anthropic.types.message_param import MessageParam
from anthropic.types.metadata_param import MetadataParam
from anthropic.types.model_param import ModelParam
from anthropic.types.text_block_param import TextBlockParam
from anthropic.types.tool_choice_param import ToolChoiceParam
from anthropic.types.tool_param import ToolParam

from ...scribe import scribe
from ..logger import (
    Generation,
    GenerationConfig,
    Logger,
    Trace,
    TraceConfig,
)
from .stream_manager import StreamWrapper
from .utils import AnthropicUtils


class MaximAnthropicMessages(Messages):
    def __init__(self, client: Anthropic, logger: Logger):
        super().__init__(client)
        self._logger = logger

    def create_non_stream(self, *args, **kwargs) -> Any:
        extra_headers = kwargs.get("extra_headers", None)
        trace_id = None
        generation_name = None
        if extra_headers is not None:
            trace_id = extra_headers.get("x-maxim-trace-id", None)
            generation_name = extra_headers.get("x-maxim-generation-name", None)
        is_local_trace = trace_id is None
        final_trace_id = trace_id or str(uuid4())
        generation: Optional[Generation] = None
        trace: Optional[Trace] = None

        try:
            openai_style_messages = None
            system = kwargs.get("system", None)
            messages = kwargs.get("messages", None)
            model = kwargs.get("model", None)
            if system is not None:
                openai_style_messages = [{"role": "system", "content": system}] + list(
                    messages
                )
            trace = self._logger.trace(TraceConfig(id=final_trace_id))
            gen_config = GenerationConfig(
                id=str(uuid4()),
                model=model,
                provider="anthropic",
                name=generation_name,
                model_parameters=AnthropicUtils.get_model_params(
                    **{
                        k: v
                        for k, v in kwargs.items()
                        if k not in ["messages", "system"]
                    }
                ),
                messages=AnthropicUtils.parse_message_param(
                    openai_style_messages
                    if openai_style_messages is not None
                    else messages  # type:ignore
                ),
            )
            generation = trace.generation(gen_config)
        except Exception as e:
            scribe().warning(
                f"[MaximSDK][AnthropicClient] Error in generating content: {str(e)}"
            )

        response = super().create(*args, **kwargs)

        try:
            if generation is not None:
                generation.result(response)
            if is_local_trace and trace is not None:
                if response is not None:
                    trace.set_output(str(response.content))
                trace.end()
        except Exception as e:
            scribe().warning(
                f"[MaximSDK][AnthropicClient] Error in logging generation: {str(e)}"
            )

        return response

    def create_stream(self, *args, **kwargs) -> Any:
        extra_headers = kwargs.get("extra_headers", None)
        trace_id = None
        generation_name = None
        if extra_headers is not None:
            trace_id = extra_headers.get("x-maxim-trace-id", None)
            generation_name = extra_headers.get("x-maxim-generation-name", None)
        is_local_trace = trace_id is None
        final_trace_id = trace_id or str(uuid4())
        generation: Optional[Generation] = None
        trace: Optional[Trace] = None

        try:
            openai_style_messages = None
            system = kwargs.get("system", None)
            messages = kwargs.get("messages", None)
            model = kwargs.get("model", None)
            if system is not None:
                openai_style_messages = [{"role": "system", "content": system}] + list(
                    messages
                )
            trace = self._logger.trace(TraceConfig(id=final_trace_id))
            gen_config = GenerationConfig(
                id=str(uuid4()),
                model=model,
                provider="anthropic",
                name=generation_name,
                model_parameters=AnthropicUtils.get_model_params(
                    **{
                        k: v
                        for k, v in kwargs.items()
                        if k not in ["messages", "system"]
                    }
                ),
                messages=AnthropicUtils.parse_message_param(
                    openai_style_messages
                    if openai_style_messages is not None
                    else messages  # type:ignore
                ),
            )
            generation = trace.generation(gen_config)
        except Exception as e:
            scribe().warning(
                f"[MaximSDK][AnthropicClient] Error in generating content: {str(e)}"
            )

        response = super().stream(*args, **kwargs)
        stream_wrapper = StreamWrapper(response, lambda chunk: process_chunk(chunk))

        def process_chunk(chunk: MessageStreamEvent):
            try:
                if chunk.type != "message_stop":
                    return
                message = chunk.message
                scribe().info(f"final_chunk: {message}")
                if generation is not None and message is not None:
                    scribe().info(f"final_chunk: {message}")
                    generation.result(message)
                    if is_local_trace and trace is not None:
                        trace.end()
            except Exception as e:
                scribe().warning(
                    f"[MaximSDK][AnthropicClient] Error in background stream listener: {str(e)}"
                )
            if is_local_trace and trace is not None:
                trace.end()

        return stream_wrapper

    def create(
        self,
        *args,
        max_tokens: int,
        messages: Iterable[MessageParam],
        model: ModelParam,
        metadata: MetadataParam | NotGiven = NOT_GIVEN,
        stop_sequences: List[str] | NotGiven = NOT_GIVEN,
        system: Union[str, Iterable[TextBlockParam]] | NotGiven = NOT_GIVEN,
        temperature: float | NotGiven = NOT_GIVEN,
        tool_choice: ToolChoiceParam | NotGiven = NOT_GIVEN,
        tools: Iterable[ToolParam] | NotGiven = NOT_GIVEN,
        top_k: int | NotGiven = NOT_GIVEN,
        top_p: float | NotGiven = NOT_GIVEN,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = NOT_GIVEN,
        **kwargs,
    ) -> Any:
        stream = kwargs.get("stream", False)
        # Add all parameters back to kwargs
        kwargs["max_tokens"] = max_tokens
        kwargs["messages"] = messages
        kwargs["model"] = model

        if metadata is not NOT_GIVEN:
            kwargs["metadata"] = metadata
        if stop_sequences is not NOT_GIVEN:
            kwargs["stop_sequences"] = stop_sequences
        if system is not NOT_GIVEN:
            kwargs["system"] = system
        if temperature is not NOT_GIVEN:
            kwargs["temperature"] = temperature
        if tool_choice is not NOT_GIVEN:
            kwargs["tool_choice"] = tool_choice
        if tools is not NOT_GIVEN:
            kwargs["tools"] = tools
        if top_k is not NOT_GIVEN:
            kwargs["top_k"] = top_k
        if top_p is not NOT_GIVEN:
            kwargs["top_p"] = top_p

        if extra_headers is not None:
            kwargs["extra_headers"] = extra_headers
        if extra_query is not None:
            kwargs["extra_query"] = extra_query
        if extra_body is not None:
            kwargs["extra_body"] = extra_body
        if timeout is not NOT_GIVEN:
            kwargs["timeout"] = timeout

        if stream:
            return self.create_stream(*args, **kwargs)
        else:
            return self.create_non_stream(*args, **kwargs)

    def stream(self, *args, **kwargs) -> Any:
        return self.create_stream(*args, **kwargs)



from typing import Optional
from uuid import uuid4

from openai import AsyncOpenAI
from openai.resources.chat import AsyncCompletions
from typing_extensions import override

from ...scribe import scribe
from ..logger import Generation, Logger, Trace
from .utils import OpenAIUtils


class MaximAsyncOpenAIChatCompletions(AsyncCompletions):
    def __init__(self, client: AsyncOpenAI, logger: Logger):
        super().__init__(client=client)
        self._logger = logger

    @override
    async def create(self, *args, **kwargs):
        metadata = kwargs.get("metadata", None)
        trace_id = None
        generation_name = None
        maxim_metadata = None
        if metadata is not None:
            maxim_metadata = metadata.get("maxim", None)
            if maxim_metadata is not None:
                trace_id = maxim_metadata.get("trace_id", None)
                generation_name = maxim_metadata.get("generation_name", None)
        is_local_trace = trace_id is None
        model = kwargs.get("model", None)
        final_trace_id = trace_id or str(uuid4())
        generation: Optional[Generation] = None
        trace: Optional[Trace] = None
        messages = kwargs.get("messages", None)
        try:
            trace = self._logger.trace({"id": final_trace_id})
            gen_config = {
                "id": str(uuid4()),
                "model": model,
                "provider": "openai",
                "name": generation_name,
                "model_parameters": OpenAIUtils.get_model_params(**kwargs),
                "messages": OpenAIUtils.parse_message_param(messages),
            }
            generation = trace.generation(gen_config)
        except Exception as e:
            scribe().warning(
                f"[MaximSDK][MaximAsyncOpenAIChatCompletions] Error in generating content: {str(e)}"
            )

        response = await super().create(*args, **kwargs)

        try:
            if generation is not None:
                generation.result(OpenAIUtils.parse_completion(response))
            if is_local_trace and trace is not None:
                trace.set_output(response.choices[0].message.content or "")
                trace.end()
        except Exception as e:
            scribe().warning(
                f"[MaximSDK][MaximAsyncOpenAIChatCompletions] Error in logging generation: {str(e)}"
            )

        return response



from dataclasses import dataclass
from typing import Dict, Literal, Union


class QueryRuleType(str):
    DeploymentVar = "deploymentVar"
    Tag = "tag"


@dataclass
class QueryRule():
    query: str
    operator: Literal["AND", "OR"]
    exact_match: bool
    scopes: Dict[str, str]


class QueryBuilder:
    def __init__(self):
        self.query: str = ""
        self.scopes: Dict[str, str] = {}
        self.operator: Literal["AND", "OR"] = "AND"
        self.is_exact_match: bool = False

    def and_(self) -> 'QueryBuilder':
        """
        Sets the operator for combining query rules to 'AND'.

        Returns:
            QueryBuilder: The current QueryBuilder instance for method chaining.
        """
        self.operator = "AND"
        return self

    def or_(self) -> 'QueryBuilder':
        """
        Sets the operator for combining query rules to 'OR'.

        Returns:
            QueryBuilder: The current QueryBuilder instance for method chaining.
        """
        self.operator = "OR"
        return self

    def folder(self, folderId: str) -> 'QueryBuilder':
        """
        Sets the folder scope for the query.

        Args:
            folderId (str): The ID of the folder to set as the scope.

        Returns:
            QueryBuilder: The current QueryBuilder instance for method chaining.
        """
        self.scopes["folder"] = folderId
        return self

    def exact_match(self) -> 'QueryBuilder':
        """
        Sets the exact match flag to True.

        Returns:
            QueryBuilder: The current QueryBuilder instance for method chaining.
        """
        self.is_exact_match = True
        return self

    def deployment_var(self, key: str, value: Union[str, int, bool], enforce: bool = True) -> 'QueryBuilder':
        """
        Adds a deployment variable rule to the query.

        Args:
            key (str): The key of the deployment variable.
            value (Union[str, int, bool]): The value of the deployment variable.
            enforce (bool, optional): Whether to enforce the deployment variable. Defaults to True.

        Returns:
            QueryBuilder: The current QueryBuilder instance for method chaining.
        """

        if len(self.query)>0:
            self.query += ","
        self.query += f"{'!!' if enforce else ''}{key}={value}"
        return self

    def tag(self, key: str, value: Union[str, int, bool], enforce: bool = False) -> 'QueryBuilder':
        """
        Adds a tag rule to the query.

        Args:
            key (str): The key of the tag.
            value (Union[str, int, bool]): The value of the tag.
            enforce (bool, optional): Whether to enforce the tag. Defaults to False.

        Returns:
            QueryBuilder: The current QueryBuilder instance for method chaining.
        """
        if len(self.query)>0:
            self.query += ","
        self.query += f"{'!!' if enforce else ''}{key}={value}"
        return self

    def build(self) -> QueryRule:
        """
        Builds the final query rule.

        Raises:
            ValueError: If the query is empty after trimming.

        Returns:
            QueryRule: A QueryRule instance with the built query.
        """
        if len(self.query.strip())==0:
            raise ValueError("Cannot build an empty query. Please add at least one rule (deploymentVar or tag).")
        return QueryRule(
            query=self.query,
            operator=self.operator,
            exact_match=self.is_exact_match,
            scopes=self.scopes
        )



import hashlib
import threading
from typing import Dict, List


class Semaphore:
    _semaphores: Dict[str, "Semaphore"] = {}

    def __init__(self, key: str, max_locks: int):
        self.key = self.hash(key)
        self.max_locks = max_locks
        self.current_locks = 0
        self.queue: List[threading.Event] = []
        Semaphore._semaphores[self.key] = self

    @staticmethod
    def hash(key: str) -> str:
        return hashlib.md5(key.encode()).hexdigest()

    def acquire(self):
        if self.current_locks < self.max_locks:
            self.current_locks += 1
            return
        event = threading.Event()
        self.queue.append(event)
        event.wait()

    def release(self):
        if self.current_locks > 0:
            self.current_locks -= 1
            if self.queue:
                next_event = self.queue.pop(0)
                self.current_locks += 1
                next_event.set()

    @classmethod
    def get(cls, key: str, max_locks: int) -> "Semaphore":
        hashed_key = cls.hash(key)
        if hashed_key not in cls._semaphores:
            return Semaphore(key, max_locks)
        return cls._semaphores[hashed_key]



import functools
import inspect
import traceback
from uuid import uuid4

from livekit.agents.voice.agent_activity import AgentActivity

from ...scribe import scribe
from .store import get_session_store
from .utils import start_new_turn

agent_activity_f_skip_list = []


def post_start(self: AgentActivity, *args, **kwargs):
    scribe().debug(f"[Internal][{self.__class__.__name__}] post start called")
    # Trying to get AgentSession and RealtimeSession handles
    rt_session_id = id(self._rt_session)
    session_info = get_session_store().get_session_by_agent_session_id(
        id(self._session)
    )
    if session_info is None:
        scribe().error("[MaximSDK] session info is none at realtime session start")
        return
    scribe().debug(
        f"[Internal][{self.__class__.__name__}] session info: {session_info}"
    )
    session_info["rt_session_id"] = rt_session_id
    session_info["rt_session"] = self._rt_session
    get_session_store().set_session(session_info)


def handle_interrupt(self: AgentActivity, *args, **kwargs):
    trace = get_session_store().get_current_trace_for_agent_session(id(self._session))
    if trace is None:
        scribe().error("[MaximSDK] trace is none at realtime session interrupt")
        return
    trace.event(id=str(uuid4()), name="interrupted")
    # here we will need to end the turn


def handle_input_speech_started(self: AgentActivity, *args, **kwargs):
    scribe().debug(f"[Internal][{self.__class__.__name__}] input speech started called")
    session_info = get_session_store().get_session_by_agent_session_id(
        id(self._session)
    )
    if session_info is None:
        scribe().error("[MaximSDK] session info is none at realtime session emit")
        return
    start_new_turn(session_info)


def handle_create_speech_task(self: AgentActivity, args, kwargs):
    if self.agent.session.agent_state != "listening":
        return
    scribe().debug(f"[Internal][{self.__class__.__name__}] create speech task called")
    session_info = get_session_store().get_session_by_agent_session_id(
        id(self._session)
    )
    if session_info is None:
        scribe().error("[MaximSDK] session info is none at realtime session emit")
        return
    if session_info["provider"] != "google-realtime":
        return
    # This is currently hack as Gemini does not support server side interruptions
    current_turn = session_info["current_turn"]
    if current_turn is None:
        scribe().error("[MaximSDK] current turn is none at realtime session emit")
        return
    input_buffer = None
    # Check if there is data and if its there copy it
    if len(current_turn["turn_input_audio_buffer"]) > 0:
        input_buffer = bytes(current_turn["turn_input_audio_buffer"])
    start_new_turn(session_info)
    if input_buffer is not None and len(input_buffer) > 0:
        session_info = get_session_store().get_session_by_agent_session_id(
            id(self._session)
        )
        if session_info is not None and input_buffer is not None:
            session_info["current_turn"]["turn_input_audio_buffer"] = bytes(
                input_buffer
            )
            get_session_store().set_session(session_info)


def pre_hook(self, hook_name, args, kwargs):
    try:
        if hook_name == "interrupt":
            handle_interrupt(self, args, kwargs)
        elif hook_name == "_on_input_speech_started":
            handle_input_speech_started(self, args, kwargs)
        elif hook_name == "_create_speech_task":
            handle_create_speech_task(self, args, kwargs)
        else:
            scribe().debug(
                f"[Internal][{self.__class__.__name__}] {hook_name} called; args={args}, kwargs={kwargs}"
            )
    except Exception as e:
        scribe().error(
            f"[{self.__class__.__name__}] {hook_name} failed; error={str(e)}\n{traceback.format_exc()}"
        )


def post_hook(self, result, hook_name, args, kwargs):
    try:
        if hook_name == "start":
            post_start(self, *args, **kwargs)    
        else:
            scribe().debug(
                f"[Internal][{self.__class__.__name__}] {hook_name} called; args={args}, kwargs={kwargs}"
            )    
    except Exception as e:
        scribe().error(
            f"[{self.__class__.__name__}] {hook_name} failed; error={str(e)}\n{traceback.format_exc()}"
        )


def instrument_agent_activity(orig, name):
    if name in agent_activity_f_skip_list:
        return orig

    if inspect.iscoroutinefunction(orig):

        async def async_wrapper(self, *args, **kwargs):
            pre_hook(self, name, args, kwargs)
            result = None
            try:
                result = await orig(self, *args, **kwargs)
                return result
            finally:                 
                post_hook(self, result, name, args, kwargs)            

        wrapper = async_wrapper
    else:

        def sync_wrapper(self, *args, **kwargs):
            pre_hook(self, name, args, kwargs)
            result = None
            try:
                result = orig(self, *args, **kwargs)
                return result
            except Exception:
                pass
            post_hook(self, result, name, args, kwargs)            

        wrapper = sync_wrapper
    return functools.wraps(orig)(wrapper)



import json
from dataclasses import asdict, dataclass
from datetime import datetime
from typing import Any, Dict, List, Optional, TypedDict, Union

from .prompt import Prompt


class AgentCost(TypedDict):
    input: float
    output: float
    total: float


class AgentUsage(TypedDict):
    prompt_tokens: int
    completion_tokens: int
    total_tokens: int


class AgentResponseMeta(TypedDict):
    cost: AgentCost
    usage: AgentUsage
    bound_variable_responses: Optional[dict[str, Any]]
    retrieved_context: Optional[str]


@dataclass
class AgentResponse:
    response: str
    meta: AgentResponseMeta

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> "AgentResponse":
        return AgentResponse(
            response=data["response"],
            meta=AgentResponseMeta(
                cost=AgentCost(**data["meta"]["cost"]),
                usage=AgentUsage(**data["meta"]["usage"]),
                bound_variable_responses=data["meta"].get("bound_variable_responses"),
                retrieved_context=data["meta"].get("retrieved_context"),
            ),
        )


@dataclass
class PromptNode:
    prompt: Prompt


@dataclass
class CodeBlockNode:
    code: str


@dataclass
class ApiParams:
    id: str
    key: str
    value: str


@dataclass
class ApiNode:
    api: dict[str, Any]


@dataclass
class Node:
    order: int
    content: Union[PromptNode, CodeBlockNode, ApiNode]

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> "Node":
        content_type = next(iter(set(data.keys()) - {"order"}))
        content_data = data[content_type]
        if content_type == "prompt":
            content = PromptNode(prompt=Prompt.from_dict(content_data))
        elif content_type == "code":
            content = CodeBlockNode(code=content_data)
        else:  # api
            content = ApiNode(api=content_data)
        return Node(order=data["order"], content=content)


# Note: Any changes here should be done in RunnablePromptChain as well
@dataclass
class PromptChain:
    prompt_chain_id: str
    version: int
    version_id: str
    nodes: List[Node]

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> "PromptChain":
        return PromptChain(
            prompt_chain_id=data["promptChainId"],
            version=data["version"],
            version_id=data["versionId"],
            nodes=[Node.from_dict(node) for node in data["nodes"]],
        )


@dataclass
class PromptChainVersionConfig:
    nodes: List[Node]

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> "PromptChainVersionConfig":
        return PromptChainVersionConfig(
            nodes=[Node.from_dict(node) for node in data["nodes"]]
        )


@dataclass
class PromptChainVersion:
    id: str
    version: int
    promptChainId: str
    description: Optional[str]
    config: Optional[PromptChainVersionConfig]
    createdAt: str
    updatedAt: str

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> "PromptChainVersion":
        return PromptChainVersion(
            id=data["id"],
            version=data["version"],
            promptChainId=data["promptChainId"],
            description=data.get("description"),
            config=(
                PromptChainVersionConfig.from_dict(data["config"])
                if data.get("config")
                else None
            ),
            createdAt=data["createdAt"],
            updatedAt=data["updatedAt"],
        )


@dataclass
class PromptChainRuleType:
    field: str
    value: Union[str, int, List[str], bool, None]  # adding None here
    operator: str
    valueSource: Optional[str] = None
    exactMatch: Optional[bool] = None

    @staticmethod
    def from_dict(obj: Dict[str, Any]):
        return PromptChainRuleType(
            field=obj["field"],
            value=obj["value"],
            operator=obj["operator"],
            valueSource=obj.get("valueSource", None),
            exactMatch=obj.get("exactMatch", None),
        )


@dataclass
class PromptChainRuleGroupType:
    rules: List[Union["PromptChainRuleType", "PromptChainRuleGroupType"]]
    combinator: str

    @staticmethod
    def from_dict(obj: Dict[str, Any]):
        rules = []
        for rule in obj["rules"]:
            if "rules" in rule:
                rules.append(PromptChainRuleGroupType.from_dict(rule))
            else:
                rules.append(PromptChainRuleType.from_dict(rule))
        return PromptChainRuleGroupType(rules=rules, combinator=obj["combinator"])


@dataclass
class PromptChainDeploymentRules:
    version: int
    query: Optional[PromptChainRuleGroupType] = None

    @staticmethod
    def from_dict(obj: Dict[str, Any]):
        query = obj.get("query", None)
        if query:
            query = PromptChainRuleGroupType.from_dict(query)
        return PromptChainDeploymentRules(version=obj["version"], query=query)


@dataclass
class VersionSpecificDeploymentConfig:
    id: str
    timestamp: datetime
    rules: PromptChainDeploymentRules
    isFallback: bool = False

    @staticmethod
    def from_dict(obj: Dict[str, Any]):
        rules = PromptChainDeploymentRules.from_dict(obj["rules"])
        return VersionSpecificDeploymentConfig(
            id=obj["id"],
            timestamp=obj["timestamp"],
            rules=rules,
            isFallback=obj.get("isFallback", False),
        )


@dataclass
class PromptChainVersionsAndRules:
    folderId: str
    rules: Dict[str, List[VersionSpecificDeploymentConfig]]
    versions: List[PromptChainVersion]
    fallbackVersion: Optional[PromptChainVersion]

    @staticmethod
    def from_dict(obj: Dict[str, Any]):
        rules = obj["rules"]
        # Decoding each rule
        for key in rules:
            rules[key] = [
                VersionSpecificDeploymentConfig.from_dict(rule) for rule in rules[key]
            ]
        versions = [
            PromptChainVersion.from_dict(version) for version in obj["versions"]
        ]
        fallbackVersion = obj.get("fallbackVersion", None)
        if fallbackVersion:
            fallbackVersion = PromptChainVersion.from_dict(fallbackVersion)
        return PromptChainVersionsAndRules(
            rules=rules,
            versions=versions,
            folderId=obj.get("folderId", None),
            fallbackVersion=fallbackVersion,
        )


@dataclass
class VersionAndRulesWithPromptChainId(PromptChainVersionsAndRules):
    promptChainId: str = ""

    @staticmethod
    def from_dict(obj: Dict[str, Any]):
        existing_rules = obj["rules"]
        rules: Dict[str, List[VersionSpecificDeploymentConfig]] = {}
        # Decoding each rule
        for key in existing_rules:
            configs: List[VersionSpecificDeploymentConfig] = []
            for rule in existing_rules[key]:
                configs.append(VersionSpecificDeploymentConfig.from_dict(rule))
            rules[key] = configs
        versions: List[PromptChainVersion] = []
        for version_dict in obj["versions"]:
            versions.append(PromptChainVersion.from_dict(version_dict))
        fallback_version: Optional[PromptChainVersion] = None

        if (fallback_version_dict := obj.get("fallbackVersion", None)) is not None:
            fallback_version = PromptChainVersion.from_dict(fallback_version_dict)
        return VersionAndRulesWithPromptChainId(
            rules=rules,
            versions=versions,
            promptChainId=obj["promptChainId"],
            folderId=obj.get("folderId", None),
            fallbackVersion=fallback_version,
        )


@dataclass
class MaximApiPromptChainResponse:
    data: PromptChainVersionsAndRules
    error: Optional[dict[str, Any]]

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> "MaximApiPromptChainResponse":
        return MaximApiPromptChainResponse(
            data=PromptChainVersionsAndRules.from_dict(data["data"]),
            error=data.get("error"),
        )


@dataclass
class PromptChainWithId(PromptChainVersionsAndRules):
    promptChainId: str

    @staticmethod
    def from_dict(obj: Dict[str, Any]):
        rules = obj["rules"]
        # Decoding each rule
        for key in rules:
            rules[key] = [
                VersionSpecificDeploymentConfig.from_dict(rule) for rule in rules[key]
            ]
        versions = [
            PromptChainVersion.from_dict(version) for version in obj["versions"]
        ]
        fallbackVersion = obj.get("fallbackVersion", None)
        if fallbackVersion:
            fallbackVersion = PromptChainVersion.from_dict(fallbackVersion)
        return PromptChainWithId(
            promptChainId=obj["promptChainId"],
            rules=rules,
            versions=versions,
            folderId=obj.get("folderId", None),
            fallbackVersion=fallbackVersion,
        )


class VersionAndRulesWithPromptChainIdEncoder(json.JSONEncoder):
    def default(self, o):
        if isinstance(o, VersionAndRulesWithPromptChainId):
            return asdict(o)
        return super().default(o)


@dataclass
class MaximApiPromptChainsResponse:
    data: List[PromptChainWithId]
    error: Optional[dict[str, Any]]

    @staticmethod
    def from_dict(incoming_data: Dict[str, Any]) -> "MaximApiPromptChainsResponse":
        return MaximApiPromptChainsResponse(
            data=[PromptChainWithId.from_dict(item) for item in incoming_data["data"]],
            error=incoming_data.get("error"),
        )



from dataclasses import dataclass
from typing import Any, Callable, Dict, List, Literal, Optional, TypeVar, Union


class VariableType(str):
    TEXT = "text"
    JSON = "json"


@dataclass
class DatasetRow:
    id: str
    data: Dict[str, str]

    def __json__(self):
        return {"id": self.id, "data": self.data}

    def to_dict(self) -> Dict[str, Any]:
        return {"id": self.id, "data": self.data}

    @classmethod
    def dict_to_class(cls, data: Dict[str, Any]) -> "DatasetRow":
        return cls(id=data["id"], data=data["data"])


class Variable:
    def __init__(self, type: str, payload: Dict[str, Union[str, int, bool]]):
        self.type = type
        self.payload = payload

    def to_json(self):
        return {"type": self.type, "payload": self.payload}

    @classmethod
    def from_json(cls, data: Dict[str, Any]) -> "Variable":
        return cls(type=data["type"], payload=data["payload"])


class DatasetEntry:
    def __init__(
        self,
        input: Variable,
        context: Optional[Variable] = None,
        expectedOutput: Optional[Variable] = None,
    ):
        self.input = input
        self.context = context
        self.expectedOutput = expectedOutput

    def to_json(self):
        return_dict = {}
        if self.input is not None:
            return_dict["input"] = {
                "type": self.input["type"],  # type: ignore
                "payload": self.input["payload"],  # type: ignore
            }
        if self.context is not None:
            return_dict["context"] = {
                "type": self.context["type"],  # type: ignore
                "payload": self.context["payload"],  # type: ignore
            }
        if self.expectedOutput is not None:
            return_dict["expectedOutput"] = {
                "type": self.expectedOutput["type"],  # type: ignore
                "payload": self.expectedOutput["payload"],  # type: ignore
            }
        return return_dict


InputColumn = Literal["INPUT"]
ExpectedOutputColumn = Literal["EXPECTED_OUTPUT"]
ContextToEvaluateColumn = Literal["CONTEXT_TO_EVALUATE"]
VariableColumn = Literal["VARIABLE"]
FileURLVariableColumn = Literal["FILE_URL_VARIABLE"]
NullableVariableColumn = Literal["NULLABLE_VARIABLE"]
OutputColumn = Literal["OUTPUT"]

DataStructure = Dict[
    str,
    Union[
        InputColumn,
        ExpectedOutputColumn,
        ContextToEvaluateColumn,
        VariableColumn,
        FileURLVariableColumn,
        NullableVariableColumn,
    ],
]

T = TypeVar("T", bound=DataStructure)

DataValue = list[T]

LocalData = Dict[str, Union[str, List[str], None]]
Data = Union[str, List[LocalData], LocalData, Callable[[int], Optional[LocalData]]]



from .attachment import SignedURLResponse
from .dataset import DatasetEntry, DatasetRow, DataStructure, Variable, VariableType
from .evaluator import (
    Evaluator,
    EvaluatorType,
    LocalEvaluatorResultParameter,
    LocalEvaluatorReturn,
    PassFailCriteria,
)
from .folder import Folder, FolderEncoder
from .prompt import (
    ChatCompletionMessage,
    ImageUrls,
    MaximApiPromptResponse,
    MaximApiPromptsResponse,
    MaximAPIResponse,
    Message,
    Prompt,
    PromptData,
    PromptResponse,
    PromptVersion,
    RuleGroupType,
    RuleType,
    Tool,
    VersionAndRulesWithPromptId,
    VersionAndRulesWithPromptIdEncoder,
    VersionsAndRules,
)
from .prompt_chain import (
    AgentResponse,
    ApiNode,
    ApiParams,
    CodeBlockNode,
    Node,
    PromptChain,
    PromptChainDeploymentRules,
    PromptChainRuleGroupType,
    PromptChainRuleType,
    PromptChainVersion,
    PromptChainVersionConfig,
    PromptChainVersionsAndRules,
    PromptChainWithId,
    PromptNode,
    VersionAndRulesWithPromptChainId,
    VersionAndRulesWithPromptChainIdEncoder,
)
from .query_builder import QueryBuilder, QueryRule, QueryRuleType
from .test_run import (
    ConsoleLogger,
    Data,
    EvaluatorConfig,
    ExecutePromptForDataResponse,
    ExecuteWorkflowForDataResponse,
    HumanEvaluationConfig,
    LocalData,
    RunResult,
    RunStatus,
    RunType,
    T,
    TestRun,
    TestRunConfig,
    TestRunEntry,
    TestRunLogger,
    TestRunResult,
    TestRunStatus,
    TestRunWithDatasetEntry,
    YieldedOutput,
    YieldedOutputMeta,
    YieldedOutputCost,
    YieldedOutputTokenUsage,
)

__all__ = [
    "PromptResponse",
    "LocalEvaluatorResultParameter",
    "LocalEvaluatorReturn",
    "DatasetRow",
    "ChatCompletionMessage",
    "Tool",
    "ImageUrls",
    "PassFailCriteria",
    "Prompt",
    "Folder",
    "PromptChain",
    "Prompt",
    "PromptVersion",
    "QueryBuilder",
    "Node",
    "Variable",
    "QueryRule",
    "PromptNode",
    "QueryRuleType",
    "RuleType",
    "ExecuteWorkflowForDataResponse",
    "ExecutePromptForDataResponse",
    "RunStatus",
    "VersionAndRulesWithPromptChainIdEncoder",
    "PromptChainVersion",
    "VariableType",
    "RuleGroupType",
    "PromptChainRuleType",
    "Data",
    "PromptChainRuleGroupType",
    "PromptChainDeploymentRules",
    "PromptChainVersionConfig",
    "PromptChainVersionsAndRules",
    "PromptChainWithId",
    "DataStructure",
    "ApiNode",
    "ApiParams",
    "CodeBlockNode",
    "VersionAndRulesWithPromptId",
    "MaximApiPromptResponse",
    "MaximAPIResponse",
    "MaximApiPromptsResponse",
    "PromptData",
    "Message",
    "VersionAndRulesWithPromptIdEncoder",
    "VersionsAndRules",
    "FolderEncoder",
    "DatasetEntry",
    "HumanEvaluationConfig",
    "VersionAndRulesWithPromptChainId",
    "EvaluatorConfig",
    "TestRun",
    "TestRunConfig",
    "Evaluator",
    "EvaluatorType",
    "RunType",
    "TestRunResult",
    "YieldedOutput",
    "SignedURLResponse",
    "LocalData",
    "T",
    "TestRunEntry",
    "TestRunStatus",
    "TestRunWithDatasetEntry",
    "YieldedOutputMeta",
    "YieldedOutputCost",
    "YieldedOutputTokenUsage",
    "TestRunLogger",
    "ConsoleLogger",
    "RunResult",
    "AgentResponse",
]



import collections.abc
import inspect
import io
import re
import struct
import types
import wave
from datetime import datetime
from decimal import Decimal
from typing import Any

from ..scribe import scribe


def is_silence(pcm_bytes, threshold=500, min_silence_ratio=0.95):
    """
    Detects if the given PCM16 byte buffer is mostly silence.

    Args:
        pcm_bytes (bytes): PCM16LE audio data.
        threshold (int): Max absolute value to consider as silence.
        min_silence_ratio (float): Minimum ratio of silent samples to consider the buffer as silence.

    Returns:
        bool: True if buffer is mostly silence, False otherwise.
    """
    num_samples = len(pcm_bytes) // 2
    if num_samples == 0:
        return True  # Empty buffer is considered silence

    silent_count = 0

    for i in range(num_samples):
        # '<h' is little-endian 16-bit signed integer
        sample = struct.unpack_from("<h", pcm_bytes, i * 2)[0]
        if abs(sample) < threshold:
            silent_count += 1

    silence_ratio = silent_count / num_samples
    return silence_ratio >= min_silence_ratio


def pcm16_to_wav_bytes(
    pcm_bytes: bytes, num_channels: int = 1, sample_rate: int = 24000
) -> bytes:
    """
    Convert PCM-16 audio data to WAV format bytes.

    Args:
        pcm_bytes (bytes): Raw PCM-16 audio data
        num_channels (int): Number of audio channels (default: 1)
        sample_rate (int): Sample rate in Hz (default: 24000)

    Returns:
        bytes: WAV format audio data
    """
    try:
        buffer = io.BytesIO()
        with wave.open(buffer, "wb") as wav_file:
            wav_file.setnchannels(num_channels)
            wav_file.setsampwidth(2)  # 16-bit PCM = 2 bytes
            wav_file.setframerate(sample_rate)
            wav_file.writeframes(pcm_bytes)
        return buffer.getvalue()
    except Exception as e:
        scribe().error(
            f"[MaximSDK] Error converting PCM-16 audio data to WAV format: {e}"
        )
        return pcm_bytes


def make_object_serializable(obj: Any) -> Any:
    """
    Convert any Python object into a JSON-serializable format while preserving
    as much information as possible about the original object.

    Args:
        obj: Any Python object

    Returns:
        A JSON-serializable representation of the object
    """
    # Handle None
    if obj is None:
        return None

    # Handle basic types that are already serializable
    if isinstance(obj, (bool, int, float, str)):
        return obj

    # Handle Decimal
    if isinstance(obj, Decimal):
        return str(obj)

    # Handle complex numbers
    if isinstance(obj, complex):
        return {"type": "complex", "real": obj.real, "imag": obj.imag}

    # Handle bytes and bytearray
    if isinstance(obj, (bytes, bytearray)):
        return {"type": "bytes", "data": obj.hex(), "encoding": "hex"}

    # Handle datetime objects
    if isinstance(obj, datetime):
        return obj.isoformat()

    # Handle regular expressions
    if isinstance(obj, re.Pattern):
        return {"type": "regex", "pattern": obj.pattern, "flags": obj.flags}

    # Handle functions
    if isinstance(obj, (types.FunctionType, types.MethodType)):
        return {
            "type": "function",
            "name": obj.__name__,
            "module": obj.__module__,
            "source": inspect.getsource(obj) if inspect.isroutine(obj) else None,
            "signature": str(inspect.signature(obj))
            if inspect.isroutine(obj)
            else None,
        }

    # Handle exceptions
    if isinstance(obj, Exception):
        return {
            "type": "error",
            "error_type": obj.__class__.__name__,
            "message": str(obj),
            "args": make_object_serializable(obj.args),
            "traceback": str(obj.__traceback__) if obj.__traceback__ else None,
        }

    # Handle sets
    if isinstance(obj, (set, frozenset)):
        return {
            "type": "set",
            "is_frozen": isinstance(obj, frozenset),
            "values": [make_object_serializable(item) for item in obj],
        }

    # Handle dictionaries and mapping types
    if isinstance(obj, collections.abc.Mapping):
        return {str(key): make_object_serializable(value) for key, value in obj.items()}

    # Handle lists, tuples, and other iterables
    if isinstance(obj, (list, tuple)) or (
        isinstance(obj, collections.abc.Iterable)
        and not isinstance(obj, (str, bytes, bytearray))
    ):
        return [make_object_serializable(item) for item in obj]

    # Handle custom objects
    try:
        # Try to convert object's dict representation
        obj_dict = obj.__dict__
        return {
            "type": "custom_object",
            "class": obj.__class__.__name__,
            "module": obj.__class__.__module__,
            "attributes": make_object_serializable(obj_dict),
        }
    except AttributeError:
        # If object doesn't have __dict__, try to get string representation
        return {
            "type": "unknown",
            "class": obj.__class__.__name__,
            "string_repr": str(obj),
        }



from .components import (
    Error,
    ErrorConfig,
    Feedback,
    FileAttachment,
    FileDataAttachment,
    Generation,
    GenerationConfig,
    GenerationConfigDict,
    GenerationError,
    GenerationRequestImageMessageContent,
    GenerationRequestMessage,
    GenerationRequestTextMessageContent,
    GenerationResult,
    GenerationResultChoice,
    GenerationUsage,
    Retrieval,
    RetrievalConfig,
    Session,
    SessionConfig,
    Span,
    SpanConfig,
    SpanConfigDict,
    ToolCall,
    ToolCallConfig,
    ToolCallError,
    Trace,
    TraceConfig,
    UrlAttachment,
    generation_request_from_gemini_content,
)
from .logger import Logger, LoggerConfig, LoggerConfigDict, get_logger_config_dict

__all__ = [
    "FileAttachment",
    "FileDataAttachment",
    "UrlAttachment",
    "Logger",
    "Error",
    "ErrorConfig",
    "LoggerConfig",
    "SpanConfigDict",
    "GenerationConfigDict",
    "GenerationResultChoice",
    "GenerationResult",
    "TraceConfig",
    "GenerationUsage",
    "GenerationRequestTextMessageContent",
    "GenerationRequestImageMessageContent",
    "Trace",
    "RetrievalConfig",
    "generation_request_from_gemini_content",
    "GenerationRequestMessage",
    "ToolCallError",
    "Retrieval",
    "SpanConfig",
    "Span",
    "GenerationConfig",
    "Generation",
    "SessionConfig",
    "Session",
    "Feedback",
    "ToolCallConfig",
    "ToolCall",
    "GenerationError",
    "LoggerConfigDict",
    "get_logger_config_dict",
]



import functools
import inspect
import traceback
import uuid
import weakref
from datetime import datetime, timezone

from livekit.agents import Agent, AgentSession
from livekit.protocol.models import Room

from ...scribe import scribe
from .store import (
    SessionState,
    SessionStoreEntry,
    Turn,
    get_livekit_callback,
    get_maxim_logger,
    get_session_store,
)


def intercept_session_start(self: AgentSession, *args, **kwargs):
    """
    This function is called when a session starts.
    This is the point where we create a new session for Maxim.
    The session info along with room_id, agent_id, etc is stored in the thread-local store.
    """
    maxim_logger = get_maxim_logger()
    scribe().debug(
        f"[Internal][{self.__class__.__name__}] Session started; args={args}, kwargs={kwargs}"
    )
    # getting the room_id
    room = kwargs.get("room", None)
    if isinstance(room, str):
        room_id = room
    elif isinstance(room, Room):
        room_id = room.sid
    else:
        room_id = id(room)
    agent: Agent = kwargs.get("agent", None)
    scribe().debug(f"[Internal]session key:{id(self)}")
    scribe().debug(f"[Internal]Room: {room_id}")
    scribe().debug(f"[Internal]Agent: {agent.instructions}")
    # creating trace as well
    session_id = str(uuid.uuid4())
    session = maxim_logger.session({"id": session_id, "name": "livekit-session"})
    # If callback is set, emit the session started event
    if get_livekit_callback() is not None:
        get_livekit_callback()(
            "maxim.session.started", {"session_id": session_id, "session": session}
        )    
    trace_id = str(uuid.uuid4())
    trace = session.trace(
        {
            "id": trace_id,
            "input": "",
            "name": "Greeting turn",
            "session_id": session_id,
            "tags": {
                "room_id": room_id,
                "agent_id": str(id(agent)),
            },
        }
    )
    if get_livekit_callback() is not None:
        get_livekit_callback()(
            "maxim.trace.started", {"trace_id": trace_id, "trace": trace}
        )
    current_turn = Turn(
        turn_id=str(uuid.uuid4()),
        turn_sequence=0,
        turn_timestamp=datetime.now(timezone.utc),
        is_interrupted=False,
        turn_input_transcription="",
        turn_output_transcription="",
        turn_input_audio_buffer=bytes(),
        turn_output_audio_buffer=bytes(),
    )
    get_session_store().set_session(
        SessionStoreEntry(
            room_id=room_id,
            user_speaking=False,
            provider="unknown",
            conversation_buffer=bytes(),
            conversation_buffer_index=1,
            state=SessionState.INITIALIZED,
            agent_id=id(agent),
            agent_session_id=id(self),
            agent_session=weakref.ref(self),
            rt_session_id=None,
            rt_session=None,
            llm_config=None,
            rt_session_info={},
            mx_current_trace_id=trace_id,
            mx_session_id=session_id,
            current_turn=current_turn,
        ),
    )


def intercept_update_agent_state(self, *args, **kwargs):
    """
    This function is called when the agent state is updated.
    """
    if not args or len(args) == 0:
        return
    new_state = args[0]
    if new_state is None:
        return
    scribe().debug(
        f"[Internal][{self.__class__.__name__}] Agent state updated; new_state={new_state}"
    )    
    trace = get_session_store().get_current_trace_for_agent_session(id(self))
    if trace is not None:
        trace.event(str(uuid.uuid4()), "agent_state_updated", {"new_state": new_state})


def intercept_generate_reply(self, *args, **kwargs):
    """
    This function is called when the agent generates a reply.
    """
    instructions = kwargs.get("instructions", None)
    if instructions is None:
        return
    scribe().debug(
        f"[Internal][{self.__class__.__name__}] Generate reply; instructions={instructions} kwargs={kwargs}"
    )
    trace = get_session_store().get_current_trace_for_agent_session(id(self))
    if trace is not None:
        trace.set_input(instructions)


def intercept_user_state_changed(self, *args, **kwargs):
    """
    This function is called when the user state is changed.
    """
    if not args or len(args) == 0:
        return
    new_state = args[0]
    if new_state is None:
        return
    scribe().debug(
        f"[Internal][{self.__class__.__name__}] User state changed; new_state={new_state}"
    )
    trace = get_session_store().get_current_trace_for_agent_session(id(self))
    if trace is not None:
        trace.event(str(uuid.uuid4()), "user_state_changed", {"new_state": new_state})


def pre_hook(self, hook_name, args, kwargs):
    try:
        if hook_name == "start":
            intercept_session_start(self, *args, **kwargs)
        elif hook_name == "_update_agent_state":
            intercept_update_agent_state(self, *args, **kwargs)
        elif hook_name == "generate_reply":
            intercept_generate_reply(self, *args, **kwargs)
        elif hook_name == "_update_user_state":
            intercept_user_state_changed(self, *args, **kwargs)
        elif hook_name == "emit":
            if args[0] == "metrics_collected":
                pass
            scribe().debug(f"[Internal][{self.__class__.__name__}] emit called; args={args}, kwargs={kwargs}")
        elif hook_name == "end":
            pass
        else:
            scribe().debug(
                f"[Internal][{self.__class__.__name__}] {hook_name} called; args={args}, kwargs={kwargs}"
            )
    except Exception as e:
        scribe().error(
            f"[{self.__class__.__name__}] {hook_name} failed; error={str(e)}\n{traceback.format_exc()}"
        )


def post_hook(self, result, hook_name, args, kwargs):
    try:
        if hook_name == "emit":
            if args[0] == "metrics_collected":
                pass
        else:
            scribe().debug(
                f"[Internal][{self.__class__.__name__}] {hook_name} completed; result={result}"
            )
    except Exception as e:
        scribe().error(
            f"[{self.__class__.__name__}] {hook_name} failed; error={str(e)}\n{traceback.format_exc()}"
        )


def instrument_agent_session(orig, name):
    if inspect.iscoroutinefunction(orig):

        async def async_wrapper(self, *args, **kwargs):
            pre_hook(self, name, args, kwargs)
            result = None
            try:
                result = await orig(self, *args, **kwargs)
                return result
            finally:
                post_hook(self, result, name, args, kwargs)            

        wrapper = async_wrapper
    else:

        def sync_wrapper(self, *args, **kwargs):
            pre_hook(self, name, args, kwargs)
            result = None
            try:
                result = orig(self, *args, **kwargs)
                return result
            finally:
                post_hook(self, result, name, args, kwargs)            

        wrapper = sync_wrapper
    return functools.wraps(orig)(wrapper)



from livekit.agents import Agent, AgentSession, JobContext, Worker
from livekit.agents.llm import RealtimeModel, RealtimeSession
from livekit.agents.voice.agent_activity import AgentActivity
from livekit.rtc import Room

from ...logger import Logger
from .agent_activity import instrument_agent_activity
from .agent_session import instrument_agent_session
from .gemini.instrumenter import instrument_gemini
from .job_context import instrument_job_context
from .realtime_session import instrument_realtime_session
from .store import MaximLiveKitCallback, set_livekit_callback, set_maxim_logger
from .worker import instrument_worker


def instrument_livekit(logger: Logger, callback: MaximLiveKitCallback = None):
    """Instrument LiveKit classes with logging.

    This function adds logging instrumentation to LiveKit classes (Agent, JobContext, LLM)
    by wrapping their methods with logging decorators. It logs method calls with their
    arguments and keyword arguments.

    The instrumentation:
    1. Wraps all Agent methods starting with "on_"
    2. Wraps all JobContext methods (except special methods)
    3. Wraps all LLM methods (except special methods)
    """
    print(
        "[MaximSDK] Warning: LiveKit instrumentation is in alpha state. Please report any issues here: https://github.com/maximhq/maxim-py/issues"
    )
    set_maxim_logger(logger)
    set_livekit_callback(callback)
    # Instrument AgentSession methods
    for name, orig in [
        (n, getattr(AgentSession, n))
        for n in dir(AgentSession)
        if callable(getattr(AgentSession, n))
    ]:
        if name != "__class__" and not name.startswith("__"):
            setattr(AgentSession, name, instrument_agent_session(orig, name))

    # Instrument Worker methods
    for name, orig in [
        (n, getattr(Worker, n)) for n in dir(Worker) if callable(getattr(Worker, n))
    ]:
        if name != "__class__" and not name.startswith("__"):
            setattr(Worker, name, instrument_worker(orig, name))

    # Instrument RealtimeSession methods
    for name, orig in [
        (n, getattr(RealtimeSession, n))
        for n in dir(RealtimeSession)
        if callable(getattr(RealtimeSession, n))
    ]:
        if name != "__class__" and not name.startswith("__"):
            setattr(RealtimeSession, name, instrument_realtime_session(orig, name))

    # Instrument AgentActivity methods
    for name, orig in [
        (n, getattr(AgentActivity, n))
        for n in dir(AgentActivity)
        if callable(getattr(AgentActivity, n))
    ]:
        if name != "__class__" and not name.startswith("__"):
            setattr(AgentActivity, name, instrument_agent_activity(orig, name))

    # Instrument JobContext methods
    for name, orig in [
        (n, getattr(JobContext, n))
        for n in dir(JobContext)
        if callable(getattr(JobContext, n))
    ]:
        if name != "__class__" and not name.startswith("__"):
            setattr(JobContext, name, instrument_job_context(orig, name))

    # Instrument gemini models if present
    instrument_gemini()


import logging
from typing import Any, Iterator, List, Optional, TypeVar, Union
from uuid import uuid4

from google.genai import Client
from google.genai.chats import Chat, Chats
from google.genai.models import Models
from google.genai.types import (
    Content,
    ContentListUnion,
    ContentListUnionDict,
    GenerateContentConfig,
    GenerateContentConfigOrDict,
    GenerateContentResponse,
    PartUnionDict,
)
from typing_extensions import override

from ..logger import (
    Generation,
    GenerationConfig,
    GenerationRequestMessage,
    Logger,
    Trace,
    TraceConfig,
)
from .async_client import MaximGeminiAsyncClient
from .utils import GeminiUtils

T = TypeVar("T")


class MaximGeminiChatSession(Chat):
    def __init__(
        self,
        chat: Chat,
        logger: Logger,
        trace_id: Optional[str] = None,
        is_local_trace: Optional[bool] = False,
    ):
        self._chat = chat
        self._logger = logger
        self._trace_id = trace_id
        self._is_local_trace = is_local_trace

    @override
    def send_message(
        self,
        message: Union[list[PartUnionDict], PartUnionDict],
        generation_name: Optional[str] = None,
    ) -> GenerateContentResponse:
        # Without trace_id we can't do anything
        if self._trace_id is None:
            return super().send_message(message)
        generation: Optional[Generation] = None
        try:
            config = self._chat._config
            messages: List[GenerationRequestMessage] = []
            if config is not None:
                if isinstance(config, GenerateContentConfig):
                    if config.system_instruction is not None:
                        messages.append(
                            GeminiUtils.parse_content(
                                config.system_instruction, "system"
                            )
                        )
                elif isinstance(config, dict):
                    if (
                        instruction := config.get("system_instruction", None)
                    ) is not None:
                        messages.append(
                            GeminiUtils.parse_content(instruction, "system")
                        )
            messages.extend(GeminiUtils.parse_chat_message("user", message))
            gen_config = GenerationConfig(
                id=str(uuid4()),
                model=self._model,
                provider="google",
                name=generation_name,
                model_parameters=GeminiUtils.get_model_params(config),
                messages=messages,
            )
            generation = self._logger.trace_generation(self._trace_id, gen_config)
            # Attaching history as metadata
            if self._curated_history is not None:
                generation.add_metadata({"history": self._curated_history})
        except Exception as e:
            logging.warning(f"[MaximSDK][Gemini] Error in generating content: {str(e)}")
        # Actual call will never fail
        response = super().send_message(message)
        # Actual call ends
        try:
            if generation is not None:
                generation.result(response)
            if self._is_local_trace:
                self._logger.trace_end(self._trace_id)
        except Exception as e:
            logging.warning(f"[MaximSDK][Gemini] Error in logging generation: {str(e)}")
        # Returning response
        return response

    @override
    def send_message_stream(
        self,
        message: Union[list[PartUnionDict], PartUnionDict],
        generation_name: Optional[str] = None,
    ):
        # Without trace_id we can't do anything
        if self._trace_id is None:
            return super().send_message_stream(message)
        generation: Optional[Generation] = None
        try:
            config = self._chat._config
            messages: List[GenerationRequestMessage] = []
            if config is not None:
                if isinstance(config, GenerateContentConfig):
                    if config.system_instruction is not None:
                        messages.append(
                            GeminiUtils.parse_content(
                                config.system_instruction, "system"
                            )
                        )
                elif isinstance(config, dict):
                    if (
                        instruction := config.get("system_instruction", None)
                    ) is not None:
                        messages.append(
                            GeminiUtils.parse_content(instruction, "system")
                        )
            messages.extend(GeminiUtils.parse_chat_message("user", message))
            gen_config = GenerationConfig(
                id=str(uuid4()),
                model=self._model,
                provider="google",
                name=generation_name,
                model_parameters=GeminiUtils.get_model_params(config),
                messages=messages,
            )
            generation = self._logger.trace_generation(self._trace_id, gen_config)
            # Attaching history as metadata
            if self._curated_history is not None:
                generation.add_metadata({"history": self._curated_history})
        except Exception as e:
            logging.warning(f"[MaximSDK][Gemini] Error in generating content: {str(e)}")
        # Actual call will never fail
        response = super().send_message_stream(message)
        # Actual call ends
        try:
            if generation is not None:
                generation.result(response)
            if self._is_local_trace:
                self._logger.trace_end(self._trace_id)
        except Exception as e:
            logging.warning(f"[MaximSDK][Gemini] Error in logging generation: {str(e)}")
        # Returning response
        return response

    def __getattr__(self, name: str) -> Any:
        result = getattr(self._chats, name)
        return result

    def __setattr__(self, name: str, value: Any) -> None:
        if name == "_chat":
            super().__setattr__(name, value)
        else:
            setattr(self._chats, name, value)

    def end_trace(self):
        if self._trace_id is not None and self._is_local_trace:
            self._logger.trace_end(self._trace_id)
            self._trace_id = None


class MaximGeminiChats(Chats):
    def __init__(self, chats: Chats, logger: Logger):
        self._chats = chats
        self._logger = logger
        self._trace_id = None
        self._is_local_trace = False

    @override
    def create(
        self,
        *,
        model: str,
        config: GenerateContentConfigOrDict = None,  # type: ignore
        history: Optional[list[Content]] = None,
        trace_id: Optional[str] = None,
    ) -> Chat:
        self._is_local_trace = trace_id is None
        self._trace_id = trace_id or str(uuid4())
        # we start generation here and send it back to chat session
        # every round trip of chat session will be logged in a separate trace
        chat_session = self._chats.create(model=model, config=config, history=history)
        maxim_chat_session = MaximGeminiChatSession(
            chat=chat_session,
            logger=self._logger,
            trace_id=self._trace_id,
            is_local_trace=self._is_local_trace,
        )
        return maxim_chat_session

    def __getattr__(self, name: str) -> Any:
        result = getattr(self._chats, name)
        return result

    def __setattr__(self, name: str, value: Any) -> None:
        if name == "_chats":
            super().__setattr__(name, value)
        else:
            setattr(self._chats, name, value)


class MaximGeminiModels(Models):
    def __init__(self, models: Models, logger: Logger):
        self._models = models
        self._logger = logger

    @override
    def generate_content_stream(
        self,
        *,
        model: str,
        contents: Union[ContentListUnion, ContentListUnionDict],
        config: Optional[GenerateContentConfigOrDict] = None,
        trace_id: Optional[str] = None,
        generation_name: Optional[str] = None,
    ) -> Iterator[GenerateContentResponse]:
        is_local_trace = trace_id is None
        final_trace_id = trace_id or str(uuid4())
        generation: Optional[Generation] = None
        trace: Optional[Trace] = None
        try:
            trace = self._logger.trace(TraceConfig(id=final_trace_id))
            # Checking if there is a system prompt
            messages: List[GenerationRequestMessage] = []
            if config is not None:
                if isinstance(config, GenerateContentConfig):
                    if config.system_instruction is not None:
                        messages.append(
                            GeminiUtils.parse_content(
                                config.system_instruction, "system"
                            )
                        )
                elif isinstance(config, dict):
                    if (
                        instruction := config.get("system_instruction", None)
                    ) is not None:
                        messages.append(
                            GeminiUtils.parse_content(instruction, "system")
                        )
            # Adding contents back
            if contents is not None:
                messages.extend(GeminiUtils.parse_messages(contents))

            gen_config = GenerationConfig(
                id=str(uuid4()),
                model=model,
                name=generation_name,
                provider="google",
                model_parameters=GeminiUtils.get_model_params(config),
                messages=messages,
            )
            generation = trace.generation(gen_config)
        except Exception as e:
            logging.warning(f"[MaximSDK][Gemini] Error in generating content: {str(e)}")
        # Actual call will never fail
        chunks = super().generate_content_stream(
            model=model, contents=contents, config=config
        )
        # Actual call ends
        try:
            if generation is not None:
                generation.result(chunks)
            if is_local_trace:
                if trace is not None:
                    if chunks is not None:
                        trace.set_output(" ".join([c.text or "" for c in chunks]))
                    trace.end()
        except Exception as e:
            logging.warning(f"[MaximSDK][Gemini] Error in logging generation: {str(e)}")
        # Actual response
        return chunks

    @override
    def generate_content(
        self,
        *,
        model: str,
        contents: Union[ContentListUnion, ContentListUnionDict],
        config: Optional[GenerateContentConfigOrDict] = None,
        trace_id: Optional[str] = None,
        generation_name: Optional[str] = None,
    ) -> GenerateContentResponse:
        is_local_trace = trace_id is None
        final_trace_id = trace_id or str(uuid4())
        generation: Optional[Generation] = None
        trace: Optional[Trace] = None
        try:
            trace = self._logger.trace(TraceConfig(id=final_trace_id))
            # Checking if there is a system prompt
            messages: List[GenerationRequestMessage] = []
            if config is not None:
                if isinstance(config, GenerateContentConfig):
                    if config.system_instruction is not None:
                        messages.append(
                            GeminiUtils.parse_content(
                                config.system_instruction, "system"
                            )
                        )
                elif isinstance(config, dict):
                    if (
                        instruction := config.get("system_instruction", None)
                    ) is not None:
                        messages.append(
                            GeminiUtils.parse_content(instruction, "system")
                        )
            # Adding contents back
            if contents is not None:
                messages.extend(GeminiUtils.parse_messages(contents))

            gen_config = GenerationConfig(
                id=str(uuid4()),
                model=model,
                provider="google",
                name=generation_name,
                model_parameters=GeminiUtils.get_model_params(config),
                messages=messages,
            )
            generation = trace.generation(gen_config)
        except Exception as e:
            logging.warning(f"[MaximSDK][Gemini] Error in generating content: {str(e)}")

        # Actual call will never fail
        response = self._models.generate_content(
            model=model, contents=contents, config=config
        )
        # Actual call ends
        try:
            if generation is not None:
                generation.result(response)
            if is_local_trace:
                if trace is not None:
                    trace.set_output(response.text or "")
                    trace.end()
        except Exception as e:
            logging.warning(f"[MaximSDK][Gemini] Error in logging generation: {str(e)}")
        # Actual response
        return response

    def __getattr__(self, name: str) -> Any:
        return getattr(self._models, name)

    def __setattr__(self, name: str, value: Any) -> None:
        if name == "_models":
            super().__setattr__(name, value)
        else:
            setattr(self._models, name, value)


class MaximGeminiClient(Client):
    def __init__(self, client: Client, logger: Logger):
        self._client = client
        self._logger = logger
        self._w_models = MaximGeminiModels(client.models, logger)
        self._w_chats = MaximGeminiChats(client.chats, logger)
        self._w_aio = MaximGeminiAsyncClient(client._aio, logger)

    @property
    def chats(self) -> MaximGeminiChats:
        return self._w_chats

    @property
    def aio(self) -> MaximGeminiAsyncClient:
        return self._w_aio

    @property
    def models(self) -> MaximGeminiModels:
        return self._w_models

    def __getattr__(self, name: str) -> Any:
        if name == "_models":
            return self._w_models
        elif name == "_chats":
            return self._w_chats
        elif name == "_aio":
            return self._w_aio
        result = getattr(self._client, name)
        return result

    def __setattr__(self, name: str, value: Any) -> None:
        if name == "_client":
            super().__setattr__(name, value)
        else:
            setattr(self._client, name, value)



from typing import List, Optional


class MaximCache():

    def get_all_keys(self) -> List[str]:
        return []

    def get(self, key: str) -> Optional[str]:
        pass

    def set(self, key: str, value: str) -> None:
        pass

    def delete(self, key: str) -> None:
        pass



import json
import logging
import time
from typing import Any, Callable, Dict, Iterator, List, Optional, Union
from uuid import uuid4

from google.genai.types import (
    Content,
    ContentDict,
    ContentListUnion,
    ContentListUnionDict,
    ContentUnion,
    File,
    GenerateContentConfig,
    GenerateContentConfigOrDict,
    GenerateContentResponse,
    Part,
    PartUnionDict,
    PIL_Image,
)

from ..logger import GenerationRequestMessage


def get_type_from_name(name: str) -> str:
    if name == "int":
        return "number"
    elif name == "str":
        return "string"
    elif name == "float":
        return "number"
    elif name == "bool":
        return "boolean"
    return "string"


class GeminiUtils:
    @staticmethod
    def parse_chat_message(
        role: str,
        message: Union[list[PartUnionDict], PartUnionDict],
    ) -> List[GenerationRequestMessage]:
        request_messages: List[GenerationRequestMessage] = []
        if isinstance(message, list):
            for m in message:
                if m is None:
                    continue
                if isinstance(m, dict):
                    pass
                else:
                    request_messages.append(
                        GenerationRequestMessage(
                            role=role, content=GeminiUtils.parse_part_text(m) or ""
                        )
                    )
        return request_messages

    @staticmethod
    def get_model_params(
        config: Optional[GenerateContentConfigOrDict] = None,
    ) -> Dict[str, Any]:
        model_params = {}
        if config is None:
            return model_params
        tools: Optional[Union[Any, List[Any]]] = None
        # We parse tool calls separately
        if isinstance(config, dict):
            for key, value in config.items():
                if key == "system_instruction" or key == "history":
                    continue
                elif key == "tools":
                    tools = value
                    continue
                model_params[key] = value
        elif isinstance(config, GenerateContentConfig):
            for key, value in config.dict():
                if key == "system_instruction" or key == "history":
                    continue
                elif key == "tools":
                    tools = value
                    continue
                model_params[key] = value

        if tools is not None:
            # parsing tools
            try:
                final_tools = []
                if isinstance(tools, list):
                    for tool in tools:
                        if isinstance(tool, Callable):
                            name = tool.__name__
                            description = tool.__doc__
                            args = []
                            for k, v in tool.__annotations__.items():
                                if k == "return":
                                    continue
                                args.append(
                                    {"name": k, "type": get_type_from_name(v.__name__)}
                                )
                            final_tools.append(
                                {
                                    "type": "function",
                                    "function": {
                                        "name": name,
                                        "description": description,
                                        "parameters": {
                                            "type": "object",
                                            "properties": {
                                                arg["name"]: {"type": arg["type"]}
                                                for arg in args
                                            },
                                        },
                                    },
                                }
                            )
                    model_params["tools"] = final_tools
            except Exception as e:
                logging.error(f"[MaximSDK] Error parsing tools: {e}")

        return model_params

    @staticmethod
    def parse_messages(
        contents: Union[ContentListUnion, ContentListUnionDict],
    ) -> List[GenerationRequestMessage]:
        messages: List[GenerationRequestMessage] = []
        if isinstance(contents, list):
            for content in contents:
                messages.append(GeminiUtils.parse_content(content))
        else:
            messages.append(GeminiUtils.parse_content(contents))
        return messages

    @staticmethod
    def parse_part_text(part: Any) -> Optional[str]:
        if isinstance(part, str):
            return part
        elif isinstance(part, File):
            return part.download_uri
        elif isinstance(part, Part):
            return part.text
        elif isinstance(part, PIL_Image):
            return "<IMAGE IS NOT SUPPORTED>"
        return None

    @staticmethod
    def parse_content(
        content: Optional[Union[ContentUnion, ContentDict]],
        override_role: Optional[str] = None,
    ) -> GenerationRequestMessage:
        message = GenerationRequestMessage(role=override_role or "user", content="")
        if isinstance(content, dict):
            if (
                role := content.get("role", None)
            ) is not None and override_role is None:
                message.role = role
            parts = content.get("parts", None)
            if parts is not None:
                for part in parts:
                    if (text := part.get("text", None)) is not None:
                        message.content += text
        elif isinstance(content, Content):
            if content.role is not None and override_role is None:
                if content.role == "model":
                    message.role = "system"
                else:
                    message.role = content.role
            if content.parts is not None:
                for part in content.parts:
                    if part is not None and part.text is not None:
                        message.content += part.text
        elif isinstance(content, List):
            for part in content:
                text = GeminiUtils.parse_part_text(part)
                if text is not None:
                    message.content += text
        else:
            text = GeminiUtils.parse_part_text(content)
            if text is not None:
                message.content += text

        return message

    @staticmethod
    def parse_gemini_generation_content_iterator(
        itr: Iterator[GenerateContentResponse],
    ) -> Dict[str, Any]:
        # Checking if it is correct or not
        if itr is None:
            raise ValueError("No response chunks")
        response: Dict[str, Any] = {}
        responses = []
        for chunk in itr:
            if not isinstance(chunk, GenerateContentResponse):
                raise ValueError("Invalid response chunk")
            responses.append(GeminiUtils.parse_gemini_generation_content(chunk))
        # merging these responses into one
        generation_result: Dict[str, Any] = {
            "id": str(uuid4()),
            "created": (
                responses[0].get("created", int(time.time()))
                if responses[0]
                else int(time.time())
            ),
            "choices": [],
            "usage": {
                "prompt_tokens": 0,
                "completion_tokens": 0,
                "total_tokens": 0,
            },
        }
        text = ""
        for response in responses:
            text += response["choices"][0]["message"]["content"]
            generation_result["usage"]["prompt_tokens"] += response["usage"][
                "prompt_tokens"
            ]
            generation_result["usage"]["completion_tokens"] += response["usage"][
                "completion_tokens"
            ]
            generation_result["usage"]["total_tokens"] += response["usage"][
                "total_tokens"
            ]
        # Create a top-level choice and then keep individual choice
        generation_result["choices"].append(
            {
                "index": 0,
                "message": {
                    "role": "assistant",
                    "content": text,
                },
                "logprobs": None,
                "finish_reason": responses[-1]["choices"][0]["finish_reason"],
            }
        )
        return generation_result

    @staticmethod
    def parse_gemini_generation_content(
        response: GenerateContentResponse,
    ) -> Dict[str, Any]:
        """
        Parse the generation response from Google's Generative AI API into a standardized format.

        Args:
            response (GenerateContentResponse): The raw response from the model

        Returns:
            Dict[str, Any]: A dictionary containing the parsed response with standardized fields:
                - usage: Token counts for prompt and completion
                - id: Generated UUID for the response
                - created: Unix timestamp of when response was parsed
        """
        generation_result: Dict[str, Any] = {}
        choices: List[Dict[str, Any]] = []
        tool_calls = []
        if (
            response.automatic_function_calling_history is not None
            and len(response.automatic_function_calling_history) > 1
        ):
            # we find the history of last model message
            model_tool_call: Optional[Content] = (
                response.automatic_function_calling_history[-2]
            )
            if model_tool_call is not None and model_tool_call.parts is not None:
                part = model_tool_call.parts[0]
                if part.function_call is not None:
                    tool_calls.append(
                        {
                            "type": "function",
                            "id": part.function_call.id or str(uuid4()),
                            "function": {
                                "name": part.function_call.name,
                                "arguments": json.dumps(part.function_call.args),
                            },
                        }
                    )
        if response.candidates is not None:
            for candidate in response.candidates:
                content = ""
                if (
                    candidate.content is not None
                    and candidate.content.parts is not None
                ):
                    for part in candidate.content.parts:
                        if part.text is not None:
                            content += part.text
                choices.append(
                    {
                        "index": 0,
                        "message": {
                            "role": "assistant",
                            "content": content,
                            "tool_calls": tool_calls,
                        },
                        "logprobs": (
                            candidate.logprobs_result.chosen_candidates
                            if candidate.logprobs_result is not None
                            else None
                        ),
                        "finish_reason": (
                            candidate.finish_reason.value
                            if candidate.finish_reason is not None
                            else ""
                        ),
                    }
                )
        prompt_tokens: int = 0
        completion_tokens: int = 0
        if response.usage_metadata is not None:
            prompt_tokens = response.usage_metadata.prompt_token_count or 0
            completion_tokens = response.usage_metadata.candidates_token_count or 0
        generation_result["choices"] = choices
        generation_result["usage"] = {
            "prompt_tokens": prompt_tokens,
            "completion_tokens": completion_tokens,
            "total_tokens": prompt_tokens + completion_tokens,
        }
        generation_result["id"] = str(uuid4())
        generation_result["created"] = int(time.time())
        return generation_result



import json
from typing import Any, Dict, Optional

from ..models.dataset import (
    ContextToEvaluateColumn,
    DataStructure,
    ExpectedOutputColumn,
    InputColumn,
)


def create_data_structure(data_structure: DataStructure) -> DataStructure:
    sanitize_data_structure(data_structure)
    return data_structure


def sanitize_data_structure(data_structure: Optional[DataStructure]) -> None:
    encountered_input = False
    encountered_expected_output = False
    encountered_context_to_evaluate = False
    if data_structure:
        for value in data_structure.values():
            if value == InputColumn:
                if encountered_input:
                    raise Exception(
                        "Data structure contains more than one input",
                        {"dataStructure": json.dumps(data_structure, indent=2)},
                    )
                else:
                    encountered_input = True
            elif value == ExpectedOutputColumn:
                if encountered_expected_output:
                    raise Exception(
                        "Data structure contains more than one expectedOutput",
                        {"dataStructure": json.dumps(data_structure, indent=2)},
                    )
                else:
                    encountered_expected_output = True
            elif value == ContextToEvaluateColumn:
                if encountered_context_to_evaluate:
                    raise Exception(
                        "Data structure contains more than one contextToEvaluate",
                        {"dataStructure": json.dumps(data_structure, indent=2)},
                    )
                else:
                    encountered_context_to_evaluate = True


def validate_data_structure(
    data_structure: Dict[str, Any], against_data_structure: Dict[str, Any]
) -> None:
    data_structure_keys = set(data_structure.keys())
    against_data_structure_keys = set(against_data_structure.keys())
    for key in data_structure_keys:
        if key not in against_data_structure_keys:
            raise Exception(
                f"The provided data structure contains key '{key}' which is not present in the dataset on the platform",
                {
                    "providedDataStructureKeys": list(data_structure_keys),
                    "platformDataStructureKeys": list(against_data_structure_keys),
                },
            )



import functools
import inspect
import time
import traceback
from typing import Union

from google.genai.types import (
    LiveConnectConfig,
    LiveServerContent,
    UsageMetadata,
)
from livekit.agents.llm import InputTranscriptionCompleted
from livekit.plugins.google.beta.realtime.realtime_api import RealtimeSession
from livekit.rtc import AudioFrame

from ....scribe import scribe
from ...components import (
    AudioContent,
    FileDataAttachment,
    GenerationResult,
    GenerationResultChoice,
    ImageContent,
    TextContent,
)
from ...utils import pcm16_to_wav_bytes
from ..store import (
    SessionStoreEntry,
    get_maxim_logger,
    get_session_store,
)


def handle_build_connect_config_result(
    self: RealtimeSession, result: LiveConnectConfig
):
    # here we will start the new generation
    # this is same as session started
    session_info = get_session_store().get_session_by_rt_session_id(id(self))
    if session_info is None:
        scribe().error("[MaximSDK] session info is none at realtime session emit")
        return
    session_info["provider"] = "google-realtime"
    llm_config = result.generation_config.model_dump()
    llm_config["speech_config"] = result.speech_config.model_dump()
    llm_config["model"] = self._opts.model
    # checking if there are any tools, add adding them in the llm_config
    if result.tools is not None and len(result.tools) > 0:
        llm_config["tools"] = []
        # Third party tools
        for tool in result.tools:
            for func_declaration in tool.function_declarations:
                llm_config["tools"].append(
                    {
                        "name": func_declaration.name,
                        "description": func_declaration.description,
                        "parameters": func_declaration.parameters,
                    }
                )
    # saving back llm_config
    session_info["llm_config"] = llm_config
    # saving back the session
    get_session_store().set_session(session_info)
    trace_id = session_info["mx_current_trace_id"]
    if trace_id is None:
        return
    trace = get_maxim_logger().trace({"id": trace_id})
    turn = session_info["current_turn"]
    if turn is None:
        return
    system_prompt = ""
    if result.system_instruction is not None:
        # We have to iterate through the system instructions
        for part in result.system_instruction.parts:
            system_prompt += part.text
    trace.generation(
        {
            "id": turn["turn_id"],
            "model": self._opts.model,
            "name": "LLM call",
            "provider": "google",
            "model_parameters": llm_config,
            "messages": [{"role": "system", "content": system_prompt}],
        }
    )
    session_info["user_speaking"] = False
    get_session_store().set_session(session_info)


def handle_server_content(self: RealtimeSession, content: LiveServerContent):
    session_info = get_session_store().get_session_by_rt_session_id(id(self))
    if session_info is None:
        scribe().error("[MaximSDK] session info is none at realtime session emit")
        return
    turn = session_info["current_turn"]
    if turn is None:
        return
    if content.output_transcription is not None:
        if session_info["user_speaking"]:
            session_info["user_speaking"] = False
            get_session_store().set_session(session_info)
    # adding transcription
    if content.output_transcription is not None:
        turn["turn_output_transcription"] += content.output_transcription.text
    # getting audio bytes from the payload
    if content.model_turn is not None:
        if content.model_turn.parts is not None and len(content.model_turn.parts) > 0:
            for part in content.model_turn.parts:
                turn["turn_output_audio_buffer"] += part.inline_data.data
                if (
                    len(session_info["conversation_buffer"])
                    + len(part.inline_data.data)
                    > 10 * 1024 * 1024
                ):
                    session_id = session_info["mx_session_id"]
                    index = session_info["conversation_buffer_index"]
                    get_maxim_logger().session_add_attachment(
                        session_id,
                        FileDataAttachment(
                            data=pcm16_to_wav_bytes(
                                session_info["conversation_buffer"]
                            ),
                            tags={"attach-to": "input"},
                            name=f"Conversation part {index}",
                            timestamp=int(time.time()),
                        ),
                    )
                    session_info["conversation_buffer"] = b""
                    session_info["conversation_buffer_index"] = index + 1
                session_info["conversation_buffer"] += part.inline_data.data
    session_info["current_turn"] = turn
    get_session_store().set_session(session_info)


def handle_google_input_transcription_completed(
    session_info: SessionStoreEntry, input_transcription: InputTranscriptionCompleted
):
    turn = session_info["current_turn"]
    if turn is None:
        return
    turn["turn_input_transcription"] = input_transcription.transcript
    get_session_store().set_session(session_info)


def handle_usage_metadata(self: RealtimeSession, usage: UsageMetadata):
    session_info = get_session_store().get_session_by_rt_session_id(id(self))
    if session_info is None:
        scribe().error("[MaximSDK] session info is none at realtime session emit")
        return
    turn = session_info["current_turn"]
    if turn is None:
        return
    # Writing final Generation result
    if (
        turn["turn_output_audio_buffer"] is not None
        and len(turn["turn_output_audio_buffer"]) > 0
    ):
        get_maxim_logger().generation_add_attachment(
            turn["turn_id"],
            FileDataAttachment(
                data=pcm16_to_wav_bytes(turn["turn_output_audio_buffer"]),
                tags={"attach-to": "output"},
                name="Assistant Response",
                timestamp=int(time.time()),
            ),
        )
    if (
        turn["turn_input_audio_buffer"] is not None
        and len(turn["turn_input_audio_buffer"]) > 0
    ):
        get_maxim_logger().generation_add_attachment(
            turn["turn_id"],
            FileDataAttachment(
                data=pcm16_to_wav_bytes(turn["turn_input_audio_buffer"]),
                tags={"attach-to": "input"},
                name="User Input",
                timestamp=int(time.time()),
            ),
        )
        session_info["current_turn"]["turn_input_audio_buffer"] = b""
    contents: list[Union[TextContent, ImageContent, AudioContent]] = []
    contents.append({"type": "audio", "transcript": turn["turn_output_transcription"]})
    choices: list[GenerationResultChoice] = []
    choices.append(
        {
            "index": 0,
            "logprobs": None,
            "finish_reason": "stop",
            "message": {"role": "assistant", "content": contents, "tool_calls": []},
        }
    )
    get_maxim_logger().generation_set_provider(turn["turn_id"], "google")
    # Parsing token details
    input_token_details: dict[str, int] = {}
    output_token_details: dict[str, int] = {}
    cached_token_details: dict[str, int] = {}
    if usage.prompt_tokens_details is not None:
        for detail in usage.prompt_tokens_details:
            if detail.modality == "TEXT":
                input_token_details["text_tokens"] = detail.token_count
            elif detail.modality == "AUDIO":
                input_token_details["audio_tokens"] = detail.token_count
    if usage.response_tokens_details is not None:
        for detail in usage.response_tokens_details:
            if detail.modality == "TEXT":
                output_token_details["text_tokens"] = detail.token_count
            elif detail.modality == "AUDIO":
                output_token_details["audio_tokens"] = detail.token_count
    if usage.cache_tokens_details is not None:
        for detail in usage.cache_tokens_details:
            if detail.modality == "TEXT":
                cached_token_details["text_tokens"] = detail.token_count
            elif detail.modality == "AUDIO":
                cached_token_details["audio_tokens"] = detail.token_count
    trace = get_session_store().get_current_trace_from_rt_session_id(
        session_info["rt_session_id"]
    )
    get_maxim_logger().trace_add_generation(
        session_info["mx_current_trace_id"],
        {
            "id": turn["turn_id"],
            "model": self._opts.model,
            "name": "LLM call",
            "provider": "google",
            "model_parameters": session_info["llm_config"],
            "messages": [{"role": "user", "content": turn["turn_input_transcription"]}],
        },
    )
    result: GenerationResult = {
        "id": turn["turn_id"],
        "object": "",
        "created": int(time.time()),
        "model": self._opts.model,
        "usage": {
            "completion_tokens": usage.response_token_count,
            "prompt_tokens": usage.prompt_token_count,
            "total_tokens": usage.total_token_count,
            "input_token_details": input_token_details,
            "output_token_details": output_token_details,
            "cached_token_details": cached_token_details,
        },
        "choices": choices,
    }
    # Setting up the generation
    get_maxim_logger().generation_result(turn["turn_id"], result)
    # Setting the output to the trace
    if session_info["rt_session_id"] is not None:
        trace = get_session_store().get_current_trace_from_rt_session_id(
            session_info["rt_session_id"]
        )
        if (
            trace is not None
            and len(choices) > 0
            and choices[0]["message"]["content"] is not None
            and isinstance(choices[0]["message"]["content"], list)
            and len(choices[0]["message"]["content"]) > 0
            and choices[0]["message"]["content"][0] is not None
            and "transcript" in choices[0]["message"]["content"][0]
        ):
            trace.set_output(choices[0]["message"]["content"][0]["transcript"])


def handle_push_audio(self: RealtimeSession, args, kwargs):
    session_info = get_session_store().get_session_by_rt_session_id(id(self))
    if session_info is None:
        scribe().error("[MaximSDK] session info is none at realtime session emit")
        return
    turn = session_info["current_turn"]
    if turn is None:
        return
    if not args or len(args) == 0:
        return
    audio_frame: AudioFrame = args[0]
    # This will help us skip silence before the turn starts
    if len(audio_frame.data) == 0 and len(turn["turn_input_audio_buffer"]) == 0:
        return
    # Checking if the audio buffer is mostly silence
    turn["turn_input_audio_buffer"] += memoryview(audio_frame._data).cast("B")
    session_info["conversation_buffer"] += memoryview(audio_frame._data).cast("B")
    session_info["current_turn"] = turn
    get_session_store().set_session(session_info)


ignored_hooks = ["_resample_audio", "_send_client_event", "_start_new_generation"]


def pre_hook(self: RealtimeSession, hook_name, args, kwargs):
    try:
        if hook_name in ignored_hooks:
            return
        elif hook_name == "_handle_server_content":
            handle_server_content(self, args[0])
        elif hook_name == "_handle_usage_metadata":
            handle_usage_metadata(self, args[0])
        elif hook_name == "push_audio":
            handle_push_audio(self, args, kwargs)
        elif hook_name == "_mark_current_generation_done":
            scribe().debug(
                f"[Internal][Gemini:{self.__class__.__name__}] _mark_current_generation_done called; args={args}, kwargs={kwargs}"
            )
        else:
            scribe().debug(
                f"[Internal][Gemini:{self.__class__.__name__}] {hook_name} called; args={args}, kwargs={kwargs}"
            )
    except Exception as e:
        scribe().error(
            f"[Internal][Gemini:{self.__class__.__name__}] {hook_name} failed; error={str(e)}\n{traceback.format_exc()}"
        )


def post_hook(self: RealtimeSession, result, hook_name, args, kwargs):
    try:
        if hook_name in ignored_hooks:
            return
        elif hook_name == "_resample_audio":
            pass
        elif hook_name == "push_audio":
            pass
        elif hook_name == "_send_client_event":
            pass
        elif hook_name == "_send_task":
            pass
        elif hook_name == "_start_new_generation":
            pass
        elif hook_name == "_handle_server_content":
            # this has audio data but post_hook is not required
            pass
        elif hook_name == "_handle_usage_metadata":
            # this has usage data
            pass
        elif hook_name == "_build_connect_config":
            handle_build_connect_config_result(self, result)
        else:
            scribe().debug(
                f"[Internal][Gemini:{self.__class__.__name__}] {hook_name} completed; result={result}"
            )
    except Exception as e:
        scribe().error(
            f"[Internal][Gemini:{self.__class__.__name__}] {hook_name} failed; error={str(e)}\n{traceback.format_exc()}"
        )


def instrument_gemini_session(orig, name):
    if inspect.iscoroutinefunction(orig):

        async def async_wrapper(self, *args, **kwargs):
            pre_hook(self, name, args, kwargs)
            result = None
            try:
                result = await orig(self, *args, **kwargs)
                return result
            finally:
                post_hook(self, result, name, args, kwargs)

        wrapper = async_wrapper
    else:

        def sync_wrapper(self, *args, **kwargs):
            pre_hook(self, name, args, kwargs)
            result = None
            try:
                result = orig(self, *args, **kwargs)
                return result
            finally:
                post_hook(self, result, name, args, kwargs)

        wrapper = sync_wrapper
    return functools.wraps(orig)(wrapper)



import importlib.util

from .async_client import MaximOpenAIAsyncClient
from .client import MaximOpenAIClient
from .utils import OpenAIUtils

if importlib.util.find_spec("openai") is None:
    raise ImportError(
        "The openai package is required. Please install it using pip: `pip install openai` or `uv add openai`"
    )

__all__ = ["OpenAIUtils", "MaximOpenAIAsyncClient", "MaximOpenAIClient"]



import json
from dataclasses import asdict, dataclass
from typing import Any, Dict, List, Optional, Union


@dataclass
class Folder:
    id: str
    name: str
    parent_folder_id: str
    tags: Optional[Dict[str, Union[str, int, bool, None]]] = None

    @staticmethod
    def from_dict(obj: dict[str, Any]) -> "Folder":
        return Folder(
            id=obj["id"],
            name=obj["name"],
            parent_folder_id=obj["parentFolderId"],
            tags=obj["tags"],
        )


class FolderEncoder(json.JSONEncoder):
    def default(self, o):
        if isinstance(o, Folder):
            return asdict(o)
        return super().default(o)


@dataclass
class Error:
    message: str


@dataclass
class MaximFolderResponse:
    data: Folder
    error: Optional[Error] = None


@dataclass
class MaximFoldersResponse:
    data: List[Folder]
    error: Optional[Error] = None



import json
from abc import ABC, abstractmethod
from dataclasses import dataclass
from enum import Enum
from typing import Any, Awaitable, Callable, Dict, Generic, List, Optional, Union

from ..evaluators import BaseEvaluator
from ..models.dataset import Data, LocalData, T, Variable
from .evaluator import (
    EvaluatorType,
    LocalEvaluationResultWithId,
)


@dataclass
class YieldedOutputTokenUsage:
    prompt_tokens: int
    completion_tokens: int
    total_tokens: int
    latency: Optional[float] = None

    def __json__(self):
        ret_dict: Dict[str, Union[int, float]] = {}
        if self.prompt_tokens is not None:
            ret_dict["prompt_tokens"] = self.prompt_tokens
        if self.completion_tokens is not None:
            ret_dict["completion_tokens"] = self.completion_tokens
        if self.total_tokens is not None:
            ret_dict["total_tokens"] = self.total_tokens
        if self.latency is not None:
            ret_dict["latency"] = self.latency
        return ret_dict

    def to_dict(self) -> Dict[str, Union[int, float]]:
        ret_dict: Dict[str, Union[int, float]] = {}
        if self.prompt_tokens is not None:
            ret_dict["prompt_tokens"] = self.prompt_tokens
        if self.completion_tokens is not None:
            ret_dict["completion_tokens"] = self.completion_tokens
        if self.total_tokens is not None:
            ret_dict["total_tokens"] = self.total_tokens
        if self.latency is not None:
            ret_dict["latency"] = self.latency
        return ret_dict

    @classmethod
    def from_json(cls, json_str: str):
        data = json.loads(json_str)
        return cls.dict_to_class(data)

    @classmethod
    def dict_to_class(cls, data: Dict[str, Any]):
        return cls(
            prompt_tokens=data.get("prompt_tokens", 0),
            completion_tokens=data.get("completion_tokens", 0),
            total_tokens=data.get("total_tokens", 0),
            latency=data.get("latency"),
        )


@dataclass
class YieldedOutputCost:
    input_cost: float
    output_cost: float
    total_cost: float

    def __json__(self):
        return {
            "input": self.input_cost,
            "output": self.output_cost,
            "total": self.total_cost,
        }

    def to_dict(self):
        return {
            "input": self.input_cost,
            "output": self.output_cost,
            "total": self.total_cost,
        }

    @classmethod
    def from_json(cls, json_str: str):
        data = json.loads(json_str)
        return cls.dict_to_class(data)

    @classmethod
    def dict_to_class(cls, data: Dict[str, Any]):
        return cls(
            input_cost=data["input"],
            output_cost=data["output"],
            total_cost=data["total"],
        )


@dataclass
class YieldedOutputMeta:
    entity_type: Optional[str] = None
    entity_id: Optional[str] = None
    usage: Optional[YieldedOutputTokenUsage] = None
    cost: Optional[YieldedOutputCost] = None

    def __json__(self):
        return {
            "entityType": self.entity_type,
            "entityId": self.entity_id,
            "usage": self.usage.to_dict() if self.usage else None,
            "cost": self.cost.to_dict() if self.cost else None,
        }

    def to_dict(self):
        return {
            "entityType": self.entity_type,
            "entityId": self.entity_id,
            "usage": self.usage.to_dict() if self.usage else None,
            "cost": self.cost.to_dict() if self.cost else None,
        }

    @classmethod
    def from_json(cls, json_str: str):
        data = json.loads(json_str)
        usage = data["usage"]
        cost = data["cost"]
        return cls(
            usage=(YieldedOutputTokenUsage.dict_to_class(usage) if usage else None),
            cost=(YieldedOutputCost.dict_to_class(cost) if cost else None),
        )

    @classmethod
    def dict_to_class(cls, data: Dict[str, Any]):
        usage = data.get("usage")
        cost = data.get("cost")
        return cls(
            usage=(YieldedOutputTokenUsage.dict_to_class(usage) if usage else None),
            cost=(YieldedOutputCost.dict_to_class(cost) if cost else None),
        )


@dataclass
class YieldedOutput:
    """
    Yielded output represents the output of `yieldsOutput` function.
    """

    data: str
    retrieved_context_to_evaluate: Optional[Union[str, list[str]]] = None
    meta: Optional[YieldedOutputMeta] = None


@dataclass
class EvaluatorArgs:
    output: str
    input: Optional[str] = None
    expectedOutput: Optional[str] = None
    contextToEvaluate: Optional[Union[str, List[str]]] = None


@dataclass
class HumanEvaluationConfig:
    emails: List[str]
    instructions: Optional[str] = None
    requester: Optional[str] = None

    def __json__(self):
        return {
            "emails": self.emails,
            "instructions": self.instructions,
            "requester": self.requester,
        }

    def to_dict(self):
        return {
            k: v
            for k, v in {
                "emails": self.emails,
                "instructions": self.instructions,
                "requester": self.requester,
            }.items()
            if v is not None
        }


class RunType(str, Enum):
    SINGLE = "SINGLE"
    COMPARISON = "COMPARISON"

    def __json__(self):
        return self.value

    @classmethod
    def from_string(cls, value: str) -> "RunType":
        try:
            return cls(value)
        except ValueError:
            raise ValueError(f"Invalid RunType: {value}")


@dataclass
class EvaluatorConfig:
    id: str
    name: str
    type: EvaluatorType
    builtin: bool
    reversed: Optional[bool] = False
    config: Optional[Any] = None

    def __json__(self):
        return {
            "id": self.id,
            "name": self.name,
            "type": self.type.value,
            "builtin": self.builtin,
            "reversed": self.reversed,
            "config": self.config,
        }

    def to_dict(self):
        return {
            k: v
            for k, v in {
                "id": self.id,
                "name": self.name,
                "type": self.type.value,
                "builtin": self.builtin,
                "reversed": self.reversed,
                "config": self.config,
            }.items()
            if v is not None
        }


@dataclass
class TestRun:
    id: str
    workspace_id: str
    eval_config: Dict[str, Any]
    human_evaluation_config: Optional[HumanEvaluationConfig] = None
    parent_test_run_id: Optional[str] = None

    def to_dict(self):
        return {
            k: v
            for k, v in {
                "id": self.id,
                "workspaceId": self.workspace_id,
                "evalConfig": self.eval_config,
                "humanEvaluationConfig": (
                    self.human_evaluation_config.to_dict()
                    if self.human_evaluation_config
                    else None
                ),
                "parentTestRunId": self.parent_test_run_id,
            }.items()
            if v is not None
        }

    def __json__(self):
        return {
            "id": self.id,
            "workspaceId": self.workspace_id,
            "evalConfig": self.eval_config,
            "humanEvaluationConfig": (
                self.human_evaluation_config.__json__()
                if self.human_evaluation_config
                else None
            ),
            "parentTestRunId": self.parent_test_run_id,
        }

    @classmethod
    def from_json(cls, json_str: str):
        data = json.loads(json_str)
        return cls(
            id=data["id"],
            workspace_id=data["workspaceId"],
            eval_config=data["evalConfig"],
            human_evaluation_config=(
                HumanEvaluationConfig(**data["humanEvaluationConfig"])
                if data.get("humanEvaluationConfig")
                else None
            ),
            parent_test_run_id=data.get("parentTestRunId"),
        )

    @classmethod
    def dict_to_class(cls, data: Dict[str, Any]) -> "TestRun":
        return cls(
            id=data["id"],
            workspace_id=data["workspaceId"],
            eval_config=data["evalConfig"],
            human_evaluation_config=(
                HumanEvaluationConfig(data["humanEvaluationConfig"])
                if data.get("humanEvaluationConfig")
                else None
            ),
            parent_test_run_id=data.get("parentTestRunId"),
        )


@dataclass
class TestRunEntry:
    variables: Dict[str, Variable]
    output: Optional[str] = None
    input: Optional[str] = None
    expected_output: Optional[str] = None
    context_to_evaluate: Optional[Union[str, List[str]]] = None
    local_evaluation_results: Optional[List[LocalEvaluationResultWithId]] = None

    def to_dict(self):
        result = {}
        if self.output is not None:
            result["output"] = self.output
        if self.input is not None:
            result["input"] = self.input
        if self.expected_output is not None:
            result["expectedOutput"] = self.expected_output
        if self.context_to_evaluate is not None:
            result["contextToEvaluate"] = self.context_to_evaluate
        if self.local_evaluation_results is not None:
            result["localEvaluationResults"] = (
                [
                    local_evaluation_result.to_dict()
                    for local_evaluation_result in self.local_evaluation_results
                ]
                if self.local_evaluation_results
                else None
            )
        if self.variables is not None:
            result["dataEntry"] = {
                key: variable.to_json() for key, variable in self.variables.items()
            }

        # Keeping only non None entries in result
        return {key: value for key, value in result.items() if value is not None}

    def __json__(self):
        return {
            key: value
            for key, value in {
                "output": self.output,
                "input": self.input,
                "expectedOutput": self.expected_output,
                "contextToEvaluate": self.context_to_evaluate,
                "localEvaluationResults": (
                    [
                        local_evaluation_result.to_dict()
                        for local_evaluation_result in self.local_evaluation_results
                    ]
                    if self.local_evaluation_results
                    else None
                ),
                "dataEntry": (
                    {
                        key: variable.to_json()
                        for key, variable in self.variables.items()
                    }
                    if self.variables
                    else None
                ),
            }.items()
            if value is not None
        }

    @classmethod
    def from_json(cls, json_str: str) -> "TestRunEntry":
        data = json.loads(json_str)
        return cls.dict_to_class(data)

    @classmethod
    def dict_to_class(cls, data: Dict[str, Any]) -> "TestRunEntry":
        local_evaluation_results = data.get("localEvaluationResults")
        data_entry = data.get("dataEntry", {})
        variables = {}
        for key, value in data_entry.items():
            if value is not None:
                variables[key] = Variable.from_json(value)

        return cls(
            output=data.get("output", None),
            input=data.get("input", None),
            expected_output=data.get("expectedOutput", None),
            context_to_evaluate=data.get("contextToEvaluate", None),
            variables=variables,
            local_evaluation_results=(
                [
                    LocalEvaluationResultWithId.dict_to_class(local_evaluation_result)
                    for local_evaluation_result in local_evaluation_results
                ]
                if local_evaluation_results
                else None
            ),
        )


@dataclass
class TestRunWithDatasetEntry(TestRun):
    def __init__(self, test_run: TestRun, dataset_entry_id: str, dataset_id: str):
        super().__init__(
            id=test_run.id,
            workspace_id=test_run.workspace_id,
            eval_config=test_run.eval_config,
            human_evaluation_config=test_run.human_evaluation_config,
            parent_test_run_id=test_run.parent_test_run_id,
        )
        self.dataset_entry_id = dataset_entry_id
        self.dataset_id = dataset_id

    def __json__(self):
        base_json = super().__json__()
        base_json.update(
            {
                "datasetEntryId": self.dataset_entry_id,
                "datasetId": self.dataset_id,
            }
        )
        return base_json

    def to_dict(self):
        return {
            k: v
            for k, v in {
                "datasetEntryId": self.dataset_entry_id,
                "datasetId": self.dataset_id,
                "id": self.id,
                "workspaceId": self.workspace_id,
                "evalConfig": self.eval_config,
                "humanEvaluationConfig": (
                    self.human_evaluation_config.__json__()
                    if self.human_evaluation_config
                    else None
                ),
                "parentTestRunId": self.parent_test_run_id,
            }.items()
            if v is not None
        }

    @classmethod
    def from_json(cls, json_str: str) -> "TestRunWithDatasetEntry":
        data = json.loads(json_str)
        return cls.dict_to_class(data)

    @classmethod
    def dict_to_class(cls, data: Dict[str, Any]) -> "TestRunWithDatasetEntry":
        base_dict = super().dict_to_class(data)
        test_run = TestRunWithDatasetEntry(
            base_dict, data["datasetEntryId"], data["datasetId"]
        )
        return test_run


class RunStatus(str, Enum):
    QUEUED = "QUEUED"
    RUNNING = "RUNNING"
    FAILED = "FAILED"
    COMPLETE = "COMPLETE"
    STOPPED = "STOPPED"


@dataclass
class TestRunStatus:
    total_entries: int
    running_entries: int
    queued_entries: int
    failed_entries: int
    completed_entries: int
    stopped_entries: int
    test_run_status: RunStatus

    def to_dict(self):
        return {
            "totalEntries": self.total_entries,
            "runningEntries": self.running_entries,
            "queuedEntries": self.queued_entries,
            "failedEntries": self.failed_entries,
            "completedEntries": self.completed_entries,
            "stoppedEntries": self.stopped_entries,
            "testRunStatus": self.test_run_status.value,
        }

    @classmethod
    def from_json(cls, json_str: str) -> "TestRunStatus":
        data = json.loads(json_str)
        return cls.dict_to_class(data)

    @classmethod
    def dict_to_class(cls, data: Dict[str, Any]) -> "TestRunStatus":
        return cls(
            total_entries=data["total"],
            running_entries=data["running"],
            queued_entries=data["queued"],
            failed_entries=data["failed"],
            completed_entries=data["completed"],
            stopped_entries=data["stopped"],
            test_run_status=RunStatus(data["testRunStatus"]),
        )


@dataclass
class EvaluatorMeanScore:
    score: Union[float, bool, str]
    out_of: Optional[float] = None
    is_pass: Optional[bool] = None

    def __json__(self):
        return {"score": self.score, "outOf": self.out_of, "pass": self.is_pass}

    @classmethod
    def from_json(cls, json_str: str) -> "EvaluatorMeanScore":
        data = json.loads(json_str)
        return cls.dict_to_class(data)

    @classmethod
    def dict_to_class(cls, data: Dict[str, Any]) -> "EvaluatorMeanScore":
        return cls(
            score=data["score"], out_of=data.get("outOf"), is_pass=data.get("pass")
        )


@dataclass
class TestRunTokenUsage:
    total: int
    input: int
    completion: int

    def __json__(self):
        return {
            "total": self.total,
            "input": self.input,
            "completion": self.completion,
        }

    @classmethod
    def from_json(cls, json_str: str) -> "TestRunTokenUsage":
        data = json.loads(json_str)
        return cls.dict_to_class(data)

    @classmethod
    def dict_to_class(cls, data: Dict[str, Any]) -> "TestRunTokenUsage":
        return cls(
            total=data["total"],
            input=data["input"],
            completion=data["completion"],
        )


@dataclass
class TestRunCost:
    total: float
    input: float
    completion: float

    def __json__(self):
        return {
            "total": self.total,
            "input": self.input,
            "completion": self.completion,
        }

    @classmethod
    def from_json(cls, json_str: str) -> "TestRunCost":
        data = json.loads(json_str)
        return cls.dict_to_class(data)

    @classmethod
    def dict_to_class(cls, data: Dict[str, Any]) -> "TestRunCost":
        return cls(
            total=data["total"],
            input=data["input"],
            completion=data["completion"],
        )


@dataclass
class TestRunLatency:
    min: float
    max: float
    p50: float
    p90: float
    p95: float
    p99: float
    mean: float
    standard_deviation: float
    total: float

    def __json__(self):
        return {
            "min": self.min,
            "max": self.max,
            "p50": self.p50,
            "p90": self.p90,
            "p95": self.p95,
            "p99": self.p99,
            "mean": self.mean,
            "standardDeviation": self.standard_deviation,
            "total": self.total,
        }

    @classmethod
    def from_json(cls, json_str: str) -> "TestRunLatency":
        data = json.loads(json_str)
        return cls.dict_to_class(data)

    @classmethod
    def dict_to_class(cls, data: Dict[str, Any]) -> "TestRunLatency":
        return cls(
            min=data["min"],
            max=data["max"],
            p50=data["p50"],
            p90=data["p90"],
            p95=data["p95"],
            p99=data["p99"],
            mean=data["mean"],
            standard_deviation=data["standardDeviation"],
            total=data["total"],
        )


@dataclass
class TestRunResultObj:
    """
    Object representing a result of a test run.
    """

    name: str
    individual_evaluator_mean_score: dict[str, EvaluatorMeanScore]
    usage: Optional[TestRunTokenUsage] = None
    cost: Optional[TestRunCost] = None
    latency: Optional[TestRunLatency] = None

    def __json__(self):
        return {
            "name": self.name,
            "individualEvaluatorMeanScore": {
                k: v.__json__() for k, v in self.individual_evaluator_mean_score.items()
            },
            "usage": self.usage.__json__() if self.usage else None,
            "cost": self.cost.__json__() if self.cost else None,
            "latency": self.latency.__json__() if self.latency else None,
        }

    @classmethod
    def from_json(cls, json_str: str) -> "TestRunResultObj":
        data = json.loads(json_str)
        return cls.dict_to_class(data)

    @classmethod
    def dict_to_class(cls, data: Dict[str, Any]) -> "TestRunResultObj":
        return cls(
            name=data["name"],
            individual_evaluator_mean_score={
                k: EvaluatorMeanScore.dict_to_class(v)
                for k, v in data["individualEvaluatorMeanScore"].items()
            },
            usage=(
                TestRunTokenUsage.dict_to_class(data["usage"])
                if data.get("usage")
                else None
            ),
            cost=(
                TestRunCost.dict_to_class(data["cost"]) if data.get("cost") else None
            ),
            latency=(
                TestRunLatency.dict_to_class(data["latency"])
                if data.get("latency")
                else None
            ),
        )


@dataclass
class TestRunResult:
    link: str
    result: List[TestRunResultObj]

    def __json__(self):
        return {"link": self.link, "result": [r.__json__() for r in self.result]}

    @classmethod
    def from_json(cls, json_str: str) -> "TestRunResult":
        data = json.loads(json_str)
        return cls.dict_to_class(data)

    @classmethod
    def dict_to_class(cls, data: Dict[str, Any]) -> "TestRunResult":
        return cls(
            link=data["link"],
            result=[TestRunResultObj.dict_to_class(r) for r in data["result"]],
        )


@dataclass
class RunResult:
    test_run_result: TestRunResult
    failed_entry_indices: List[int]


class TestRunLogger(ABC):
    @abstractmethod
    def info(self, message: str) -> None:
        """
        Log an informational message.

        Args:
            message (str): The message to be logged.
        """
        pass

    @abstractmethod
    def error(self, message: str, e: Optional[Exception] = None) -> None:
        """
        Log an error message.

        Args:
            message (str): The error message to be logged.
        """
        pass


class ConsoleLogger(TestRunLogger):
    def info(self, message: str) -> None:
        print(message)

    def error(self, message: str, e: Optional[Exception] = None) -> None:
        print(message, e)


@dataclass
class WorkflowConfig:
    id: str
    context_to_evaluate: Optional[str] = None

    def __json__(self):
        return {
            "id": self.id,
            "contextToEvaluate": self.context_to_evaluate,
        }

    @classmethod
    def from_json(cls, json_str: str) -> "WorkflowConfig":
        data = json.loads(json_str)
        return cls.dict_to_class(data)

    @classmethod
    def dict_to_class(cls, data: Dict[str, Any]) -> "WorkflowConfig":
        return cls(
            id=data["id"],
            context_to_evaluate=data.get("contextToEvaluate"),
        )


@dataclass
class PromptVersionConfig:
    id: str
    context_to_evaluate: Optional[str] = None

    def __json__(self):
        return {
            "id": self.id,
            "contextToEvaluate": self.context_to_evaluate,
        }

    @classmethod
    def from_json(cls, json_str: str) -> "PromptVersionConfig":
        data = json.loads(json_str)
        return cls.dict_to_class(data)

    @classmethod
    def dict_to_class(cls, data: Dict[str, Any]) -> "PromptVersionConfig":
        return cls(
            id=data["id"],
            context_to_evaluate=data.get("contextToEvaluate"),
        )


@dataclass
class PromptChainVersionConfig:
    id: str
    context_to_evaluate: Optional[str] = None

    def __json__(self):
        return {
            "id": self.id,
            "contextToEvaluate": self.context_to_evaluate,
        }

    @classmethod
    def from_json(cls, json_str: str) -> "PromptChainVersionConfig":
        data = json.loads(json_str)
        return cls.dict_to_class(data)

    @classmethod
    def dict_to_class(cls, data: Dict[str, Any]) -> "PromptChainVersionConfig":
        return cls(
            id=data["id"],
            context_to_evaluate=data.get("contextToEvaluate"),
        )


@dataclass
class TestRunConfig(Generic[T]):
    """
    Configuration for a test run.

    Attributes:
        base_url (str): The base URL for the API.
        api_key (str): The API key for authentication.
        in_workspace_id (str): The ID of the workspace.
        workflow_id (Optional[str]): The ID of the workflow.
        prompt_version_id (Optional[str]): The ID of the prompt version.
        prompt_chain_version_id (Optional[str]): The ID of the prompt chain version.
        name (str): The name of the test run.
        data_structure (Optional[T]): The structure of the test data.
        data (Optional[Union[str, DataValue[T], Callable[[int], Optional[DataValue[T]]]]]): The test data or a function to retrieve it.
        test_config_id (Optional[str]): The ID of the test configuration.
        platform_evaluators (List[PlatformEvaluatorType[T]]): List of platform evaluators to use.
    """

    base_url: str
    api_key: str
    in_workspace_id: str
    name: str
    evaluators: List[Union[str, BaseEvaluator]]
    workflow: Optional[WorkflowConfig] = None
    prompt_version: Optional[PromptVersionConfig] = None
    prompt_chain_version: Optional[PromptChainVersionConfig] = None
    data_structure: Optional[T] = None
    data: Optional[Data] = None
    test_config_id: Optional[str] = None
    logger: TestRunLogger = ConsoleLogger()
    human_evaluation_config: Optional[HumanEvaluationConfig] = None
    output_function: Optional[
        Callable[[LocalData], Union[YieldedOutput, Awaitable[YieldedOutput]]]
    ] = None
    concurrency: Optional[int] = None


@dataclass
class ExecuteWorkflowForDataResponse:
    output: Optional[str]
    context_to_evaluate: Optional[str]
    latency: float

    def __json__(self):
        return {
            "output": self.output,
            "contextToEvaluate": self.context_to_evaluate,
            "latency": self.latency,
        }

    @classmethod
    def from_json(cls, json_str: str) -> "ExecuteWorkflowForDataResponse":
        data = json.loads(json_str)
        return cls.dict_to_class(data)

    @classmethod
    def dict_to_class(cls, data: Dict[str, Any]) -> "ExecuteWorkflowForDataResponse":
        return cls(
            output=data["output"],
            context_to_evaluate=data.get("contextToEvaluate"),
            latency=data["latency"],
        )


@dataclass
class ExecutePromptForDataResponse:
    output: Optional[str]
    context_to_evaluate: Optional[str]
    usage: Optional[YieldedOutputTokenUsage] = None
    cost: Optional[YieldedOutputCost] = None

    def __json__(self):
        return {
            "output": self.output,
            "contextToEvaluate": self.context_to_evaluate,
            "usage": self.usage.__json__() if self.usage else None,
            "cost": self.cost.__json__() if self.cost else None,
        }

    @classmethod
    def from_json(cls, json_str: str) -> "ExecutePromptForDataResponse":
        data = json.loads(json_str)
        return cls.dict_to_class(data)

    @classmethod
    def dict_to_class(cls, data: Dict[str, Any]) -> "ExecutePromptForDataResponse":
        return cls(
            output=data["output"],
            context_to_evaluate=data.get("contextToEvaluate"),
            usage=(
                YieldedOutputTokenUsage.dict_to_class(data["usage"])
                if data.get("usage")
                else None
            ),
            cost=(
                YieldedOutputCost.dict_to_class(data["cost"])
                if data.get("cost")
                else None
            ),
        )


@dataclass
class ExecutePromptChainForDataResponse:
    output: Optional[str]
    context_to_evaluate: Optional[str]
    usage: Optional[YieldedOutputTokenUsage] = None
    cost: Optional[YieldedOutputCost] = None

    def __json__(self):
        return {
            "output": self.output,
            "contextToEvaluate": self.context_to_evaluate,
            "usage": self.usage.__json__() if self.usage else None,
            "cost": self.cost.__json__() if self.cost else None,
        }

    @classmethod
    def from_json(cls, json_str: str) -> "ExecutePromptChainForDataResponse":
        data = json.loads(json_str)
        return cls.dict_to_class(data)

    @classmethod
    def dict_to_class(cls, data: Dict[str, Any]) -> "ExecutePromptChainForDataResponse":
        return cls(
            output=data["output"],
            context_to_evaluate=data.get("contextToEvaluate"),
            usage=(
                YieldedOutputTokenUsage.dict_to_class(data["usage"])
                if data.get("usage")
                else None
            ),
            cost=(
                YieldedOutputCost.dict_to_class(data["cost"])
                if data.get("cost")
                else None
            ),
        )
